# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from itertools import chain
from xml.sax.saxutils import escape
from xml.sax.saxutils import quoteattr
import logging
import sys

from .filestructure import VERSION
from .dataio import typed_struct_attributes
from .dataio import Struct
from .dataio import StructType
from .dataio import ArrayType
from .dataio import FlagsType
from .dataio import EnumType
from .dataio import WCHAR
from .dataio import HWPUNIT
from .dataio import HWPUNIT16
from .dataio import SHWPUNIT
from .binmodel import COLORREF
from .binmodel import BinStorageId
from .binmodel import Margin
from .binmodel import Text
from .treeop import STARTEVENT
from .treeop import ENDEVENT


PY3 = sys.version_info.major == 3
if PY3:
    basestring = str
    unichr = chr


logger = logging.getLogger(__name__)


def xmlattrval(value):
    if isinstance(value, basestring):
        return value
    elif isinstance(value, float):
        # https://stackoverflow.com/questions/25898733/why-does-strfloat-return-more-digits-in-python-3-than-python-2
        return repr(value)
    elif isinstance(type(value), EnumType):
        return value.name.lower() if value.name else str(int(value))
    elif isinstance(value, type):
        return value.__name__
    else:
        return str(value)


def expanded_xmlattribute(ntv):
    name, (t, value) = ntv
    if isinstance(t, FlagsType):
        fmt = '%0'
        fmt += '%d' % (t.basetype.fixed_size * 2)
        fmt += 'X'
        yield name, fmt % int(value)
        for k, v in t.dictvalue(t(value)).items():
            yield k, xmlattrval(v)
    elif t is Margin:
        for pos in ('left', 'right', 'top', 'bottom'):
            yield '-'.join([name, pos]), xmlattrval(value.get(pos))
    elif t is COLORREF:
        yield name, xmlattrval(t(value))
    elif t is VERSION:
        yield name, '.'.join(str(x) for x in value)
    elif t in (HWPUNIT, SHWPUNIT, HWPUNIT16):
        yield name, str(value)
    elif t is WCHAR:
        if value == 0:
            yield name, u''
        else:
            if value in PUA_SYMBOLS:
                yield name, PUA_SYMBOLS[value]
            else:
                yield name, unichr(value)
    elif t is BinStorageId:
        yield name, 'BIN%04X' % value
    else:
        yield name, xmlattrval(value)


# TODO: arbitrary assignment; not based on any standards
PUA_SYMBOLS = {
    0xF046: u'☞',  # U+261E WHITE RIGHT POINTING INDEX
    0xF06C: u'●',  # U+25CF BLACK CIRCLE
    # F06C: u'⚫',  # U+26AB MEDIUM BLACK CIRCLE
    0xF09F: u'•',  # U+2022 BULLET = black small circle
    0xF0A1: u'○',  # U+25CB WHITE CIRCLE
    # F0A1: u'⚪',  # U+26AA MEDIUM WHITE CIRCLE
    # F0A1: u'⚬',  # U+26AC MEDIUM SMALL WHITE CIRCLE
    # F0A1: u' ',  # U+25E6 WHITE BULLET
    0xF06E: u'■',  # U+25A0 BLACK SQUARE = molding mark
    0xF0A7: u'▪',  # U+25AA BLACK SMALL SQUARE = square bullet
    0xF06F: u'☐',  # U+2610 BALLOT BOX
    # F06F: u'□',  # U+25A1 WHITE SQUARE = quadrature
    0xF075: u'◆',  # U+25C6 BLACK DIAMOND
    0xF077: u'⬩',  # U+2B29 BLACK SMALL DIAMOND
    # F077: u'⬥',  # U+2B25 BLACK MEDIUM DIAMOND
    # F077: u'⬦',  # U+2B26 WHITE MEDIUM DIAMOND
    0xF076: u'❖',  # U+2756 BLACK DIAMOND MINUS WHITE X
    0xF0A4: u'◉',  # U+25C9 FISHEYE
    # F0A4: u'⦿ ', # U+29BF CIRCLED BULLET
    0xF0AB: u'★',  # U+2605 BLACK STAR
    0xF0Fc: u'✓',  # U+2713 CHECK MARK
    0xF0FE: u'☑',  # U+2611 BALLOT BOX WITH CHECK
}


def xmlattr_dashednames(attrs):
    for k, v in attrs:
        yield k.replace('_', '-'), v


def xmlattr_uniqnames(attrs):
    names = set([])
    for k, v in attrs:
        assert k not in names, 'name clashes: %s' % k
        yield k, v
        names.add(k)


def xmlattributes_for_plainvalues(context, plainvalues):
    ntvs = plainvalues.items()
    ntvs = chain(*(expanded_xmlattribute(ntv) for ntv in ntvs))
    return dict(xmlattr_uniqnames(xmlattr_dashednames(ntvs)))


def is_complex_type(type, value):
    if isinstance(value, dict):
        return True
    elif isinstance(type, ArrayType) and issubclass(type.itemtype, Struct):
        return True
    elif isinstance(type, ArrayType) and issubclass(type.itemtype, COLORREF):
        return True
    else:
        return False


def separate_plainvalues(typed_attributes):
    d = []
    p = dict()
    for named_item in typed_attributes:
        name, item = named_item
        t, value = item
        try:
            if t is Margin:
                p[name] = item
            elif is_complex_type(t, value):
                d.append(named_item)
            else:
                p[name] = item
        except Exception as e:
            logger.error('%s', (name, t, value))
            logger.error('%s', t.__dict__)
            logger.exception(e)
            raise e
    return d, p


def startelement(context, ma):
    model, attributes = ma
    if isinstance(model, StructType):
        typed_attributes = ((v['name'], (v['type'], v['value']))
                            for v in typed_struct_attributes(model, attributes,
                                                             context))
    else:
        typed_attributes = ((k, (type(v), v))
                            for k, v in attributes.items())

    typed_attributes, plainvalues = separate_plainvalues(typed_attributes)

    if model is Text:
        text = plainvalues.pop('text')[1]
    elif '<text>' in plainvalues:
        text = plainvalues.pop('<text>')[1]
    else:
        text = None

    yield STARTEVENT, (model.__name__,
                       xmlattributes_for_plainvalues(context, plainvalues))
    if text:
        yield Text, text

    for _name, (_type, _value) in typed_attributes:
        if isinstance(_value, dict):
            assert isinstance(_value, dict)
            _value = dict(_value)
            _value['attribute-name'] = _name
            for x in element(context, (_type, _value)):
                yield x
        else:
            assert isinstance(_value, (tuple, list)), (_value, _type)
            # assert issubclass(_type.itemtype, Struct), (_value, _type)
            if issubclass(_type.itemtype, Struct):
                yield STARTEVENT, ('Array', {'name': _name})
                for _itemvalue in _value:
                    for x in element(context, (_type.itemtype, _itemvalue)):
                        yield x
                yield ENDEVENT, 'Array'
            elif issubclass(_type.itemtype, COLORREF):
                for _itemvalue in _value:
                    yield STARTEVENT, (_name, {
                        'r': '%d' % ((_itemvalue >> 0) & 0xff),
                        'g': '%d' % ((_itemvalue >> 8) & 0xff),
                        'b': '%d' % ((_itemvalue >> 16) & 0xff),
                        'alpha': '%d' % ((_itemvalue >> 24) & 0xff),
                        'hex': xmlattrval(_type.itemtype(_itemvalue))
                    })
                    yield ENDEVENT, _name
            else:
                assert False, (_value, _type)


def element(context, ma):
    model, attributes = ma
    for x in startelement(context, ma):
        yield x
    yield ENDEVENT, model.__name__


def xmlevents_to_bytechunks(xmlevents, encoding='utf-8'):
    for textchunk in xmlevents_to_textchunks(xmlevents):
        yield textchunk.encode(encoding)


def xmlevents_to_textchunks(xmlevents):
    entities = {'\r': '&#13;',
                '\n': '&#10;',
                '\t': '&#9;'}
    for event, item in xmlevents:
        if event is STARTEVENT:
            yield '<'
            yield item[0]
            for n, v in item[1].items():
                yield ' '
                yield n
                yield '='
                v = quoteattr(v, entities)
                v = v.replace('\x00', '')
                yield v
            yield '>'
        elif event is Text:
            text = escape(item)
            text = text.replace('\x00', '')
            yield text
        elif event is ENDEVENT:
            yield '</'
            yield item
            yield '>'



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_PARA_SHAPE
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import INT32
from hwp5.dataio import UINT32
from hwp5.dataio import UINT16
from hwp5.dataio import SHWPUNIT
from hwp5.dataio import HWPUNIT16


class ParaShape(RecordModel):
    ''' 4.1.10. 문단 모양 '''
    tagid = HWPTAG_PARA_SHAPE

    # 표 39 문단 모양 속성1
    LineSpacingType = Enum(RATIO=0, FIXED=1, SPACEONLY=2, MINIMUM=3)
    Align = Enum(BOTH=0, LEFT=1, RIGHT=2, CENTER=3, DISTRIBUTE=4,
                 DISTRIBUTE_SPACE=5)
    VAlign = Enum(FONT=0, TOP=1, CENTER=2, BOTTOM=3)
    LineBreakAlphabet = Enum(WORD=0, HYPHEN=1, CHAR=2)
    LineBreakHangul = Enum(WORD=0, CHAR=1)
    HeadShape = Enum(NONE=0, OUTLINE=1, NUMBER=2, BULLET=3)
    Flags1 = Flags(UINT32,
                   0, 1, LineSpacingType, 'linespacing_type',
                   2, 4, Align, 'align',
                   5, 6, LineBreakAlphabet, 'linebreak_alphabet',
                   7, LineBreakHangul, 'linebreak_hangul',
                   8, 'use_paper_grid',
                   9, 15, 'minimum_space',  # 공백 최소값
                   16, 'protect_single_line',  # 외톨이줄 보호
                   17, 'with_next_paragraph',  # 다음 문단과 함께
                   18, 'protect',  # 문단 보호
                   19, 'start_new_page',  # 문단 앞에서 항상 쪽 나눔
                   20, 21, VAlign, 'valign',
                   22, 'lineheight_along_fontsize',  # 글꼴에 어울리는 줄 높이
                   23, 24, HeadShape, 'head_shape',  # 문단 머리 모양
                   25, 27, 'level',  # 문단 수준
                   28, 'linked_border',  # 문단 테두리 연결 여부
                   29, 'ignore_margin',  # 문단 여백 무시
                   30, 'tail_shape')  # 문단 꼬리 모양

    # 표 40 문단 모양 속성2
    Flags2 = Flags(UINT32,
                   0, 1, 'in_single_line',
                   2, 3, 'reserved',
                   4, 'autospace_alphabet',
                   5, 'autospace_number')

    # 표 41 줄 간격 종류
    Flags3 = Flags(UINT32,
                   0, 4, LineSpacingType, 'linespacing_type3')

    Flags = Flags1

    def attributes(cls):
        ''' 표 38 문단 모양 '''
        yield cls.Flags, 'parashapeflags',
        yield INT32,  'doubled_margin_left',   # 1/7200 * 2 # DIFFSPEC
        yield INT32,  'doubled_margin_right',  # 1/7200 * 2
        yield SHWPUNIT,  'indent',
        yield INT32,  'doubled_margin_top',    # 1/7200 * 2
        yield INT32,  'doubled_margin_bottom',  # 1/7200 * 2
        yield SHWPUNIT,  'linespacing',
        yield UINT16, 'tabdef_id',
        yield UINT16, 'numbering_bullet_id',
        yield UINT16, 'borderfill_id',
        yield HWPUNIT16,  'border_left',
        yield HWPUNIT16,  'border_right',
        yield HWPUNIT16,  'border_top',
        yield HWPUNIT16,  'border_bottom',
        yield dict(type=cls.Flags2, name='flags2', version=(5, 0, 1, 7))
        # yield cls.Flags3, 'flags3',   # DIFFSPEC
        # yield UINT32, 'lineSpacing',  # DIFFSPEC
    attributes = classmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''HWPv5 to ODT converter

Usage::

    hwp5odt [options] [--embed-image] <hwp5file>
    hwp5odt [options] --styles <hwp5file>
    hwp5odt [options] --content [--embed-image] <hwp5file>
    hwp5odt [options] --document [--no-embed-image] <hwp5file>
    hwp5odt -h | --help
    hwp5odt --version

Options::

    -h --help           Show this screen
    --version           Show version
    --loglevel=<level>  Set log level.
    --logfile=<file>    Set log file.

    --document          Produce single OpenDocument XML file (.fodt)
    --styles            Produce *.styles.xml
    --content           Produce *.content.xml

    --output=<file>     Output file.
'''
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from contextlib import contextmanager
from contextlib import closing
from functools import partial
from io import BytesIO
import gettext
import io
import logging
import os.path
import sys

from .errors import ImplementationNotAvailable
from .utils import mkstemp_open
from .utils import hwp5_resources_path
from .transforms import BaseTransform
from .plat import get_relaxng_compile
from .utils import cached_property


PY3 = sys.version_info.major == 3
if PY3:
    basestring = str
    unicode = str
logger = logging.getLogger(__name__)
locale_dir = os.path.join(os.path.dirname(__file__), 'locale')
locale_dir = os.path.abspath(locale_dir)
t = gettext.translation('hwp5odt', locale_dir, fallback=True)
if PY3:
    _ = t.gettext
else:
    _ = t.ugettext


RESOURCE_PATH_RNG = 'odf-relaxng/OpenDocument-v1.2-os-schema.rng'
RESOURCE_PATH_XSL_SINGLE_DOCUMENT = 'xsl/odt/document.xsl'
RESOURCE_PATH_XSL_STYLE = 'xsl/odt/styles.xsl'
RESOURCE_PATH_XSL_CONTENT = 'xsl/odt/content.xsl'


class ODFValidate:

    def __init__(self, relaxng_compile=None):
        '''
        >>> V = ODFValidate()
        '''
        if relaxng_compile is None:
            try:
                relaxng_compile = self.get_default_relaxng_compile()
            except ImplementationNotAvailable:
                relaxng_compile = None
        self.relaxng_compile = relaxng_compile

    @classmethod
    def get_default_relaxng_compile(cls):
        relaxng_compile = get_relaxng_compile()
        if not relaxng_compile:
            raise ImplementationNotAvailable('relaxng')
        return relaxng_compile

    @cached_property
    def odf_validator(self):
        '''
        >>> with V.odf_validator(sys.stdout) as output:
        ...     output.write(xml)
        '''
        return self.make_odf_validator()

    def make_odf_validator(self):
        if self.relaxng_compile:
            with hwp5_resources_path(RESOURCE_PATH_RNG) as rng_path:
                return self.relaxng_compile(rng_path)


class ODTTransform(BaseTransform, ODFValidate):

    def __init__(self, xslt_compile=None, relaxng_compile=None,
                 embedbin=False):
        '''
        >>> from hwp5.hwp5odt import ODTTransform
        >>> T = ODTTransform()
        '''
        BaseTransform.__init__(self, xslt_compile=xslt_compile,
                               embedbin=embedbin)
        ODFValidate.__init__(self, relaxng_compile)

    @property
    def transform_hwp5_to_styles(self):
        '''
        >>> with io.open('styles.xml', 'wb') as f:
        ...     T.transform_hwp5_to_styles(hwp5file, f)
        '''
        transform_xhwp5 = self.transform_xhwp5_to_styles
        return self.make_transform_hwp5(transform_xhwp5)

    @property
    def transform_hwp5_to_content(self):
        '''
        >>> with io.open('content.xml', 'wb') as f:
        ...     T.transform_hwp5_to_content(hwp5file, f)
        '''
        transform_xhwp5 = self.transform_xhwp5_to_content
        return self.make_transform_hwp5(transform_xhwp5)

    @property
    def transform_hwp5_to_single_document(self):
        '''
        >>> with io.open('transformed.fodt', 'wb') as f:
        ...     T.transform_hwp5_to_single_document(hwp5file, f)
        '''
        transform_xhwp5 = self.transform_xhwp5_to_single_document
        return self.make_transform_hwp5(transform_xhwp5)

    def transform_hwp5_to_package(self, hwp5file, odtpkg):
        '''
        >>> with open_odtpkg('transformed.odt') as odtpkg:
        ...    T.transform_hwp5_to_package(hwp5file, odtpkg)
        '''
        with self.transformed_xhwp5_at_temp(hwp5file) as xml_path:
            self.transform_xhwp5_into_package(xml_path, odtpkg)

        if 'BinData' in hwp5file:
            bindata = hwp5file['BinData']
            for name in bindata:
                f = bindata[name].open()
                path = 'bindata/' + name
                mimetype = 'application/octet-stream'
                odtpkg.insert_stream(f, path, mimetype)

    @cached_property
    def transform_xhwp5_to_styles(self):
        '''
        >>> with io.open('styles.xml', 'wb') as f:
        ...     T.transform_xhwp5_to_styles('input.xml', f)
        '''
        resource_path = RESOURCE_PATH_XSL_STYLE
        return self.make_odf_transform(resource_path)

    @cached_property
    def transform_xhwp5_to_content(self):
        '''
        >>> with io.open('content.xml', 'wb') as f:
        ...     T.transform_xhwp5_to_content('input.xml', f)
        '''
        resource_path = RESOURCE_PATH_XSL_CONTENT
        return self.make_odf_transform(resource_path)

    @cached_property
    def transform_xhwp5_to_single_document(self):
        '''
        >>> with io.open('transformed.fodf', 'wb') as f:
        ...     T.transform_xhwp5_to_single_document('input.xml', f)
        '''
        resource_path = RESOURCE_PATH_XSL_SINGLE_DOCUMENT
        return self.make_odf_transform(resource_path)

    @property
    def transform_xhwp5_into_package(self):
        '''
        >>> with open_odtpkg('transformed.odt') as odtpkg:
        >>>     T.transform_xhwp5_into_package('input.xml', odtpkg)
        '''
        def transform(xhwp5path, odtpkg):
            with self.transformed_styles_at_temp(xhwp5path) as path:
                odtpkg.insert_path(path, 'styles.xml', 'text/xml')
            with self.transformed_content_at_temp(xhwp5path) as path:
                odtpkg.insert_path(path, 'content.xml', 'text/xml')

            rdf = BytesIO()
            manifest_rdf(rdf)
            rdf.seek(0)
            odtpkg.insert_stream(rdf, 'manifest.rdf',
                                 'application/rdf+xml')
        return transform

    def transformed_styles_at_temp(self, xhwp5path):
        '''
        >>> with T.transformed_styles_at_temp('input.xml') as styles_path:
        ...     pass
        '''
        transform_xhwp5 = self.transform_xhwp5_to_styles
        return transformed_at_temp_path(xhwp5path, transform_xhwp5)

    def transformed_content_at_temp(self, xhwp5path):
        '''
        >>> with T.transformed_content_at_temp('input.xml') as content_path:
        ...     pass
        '''
        transform_xhwp5 = self.transform_xhwp5_to_content
        return transformed_at_temp_path(xhwp5path, transform_xhwp5)

    def transformed_single_document_at_temp(self, xhwp5path):
        '''
        >>> with T.transformed_single_document_at_temp('input.xml') as path:
        ...     pass
        '''
        transform_xhwp5 = self.transform_xhwp5_to_single_document
        return transformed_at_temp_path(xhwp5path, transform_xhwp5)

    def make_odf_transform(self, resource_path):
        transform = self.make_xsl_transform(resource_path)
        validator = self.odf_validator
        if validator:
            def validating_transform(input, output):
                with validator.validating_output(output) as output:
                    transform(input, output)
            return validating_transform
        else:
            return transform


@contextmanager
def transformed_at_temp_path(inp_path, transform):
    with mkstemp_open() as (tmp_path, f):
        transform(inp_path, f)
        f.flush()
        yield tmp_path


class ODTPackage(object):
    def __init__(self, path_or_zipfile):
        self.files = []

        if isinstance(path_or_zipfile, basestring):
            from zipfile import ZipFile
            zipfile = ZipFile(path_or_zipfile, 'w')
        else:
            zipfile = path_or_zipfile
        self.zf = zipfile

    def insert_path(self, src_path, path, media_type):
        with io.open(src_path, 'rb') as f:
            self.insert_stream(f, path, media_type)

    def insert_stream(self, f, path, media_type):
        if not isinstance(path, unicode):
            path = path.decode('utf-8')
        self.zf.writestr(path, f.read())
        self.files.append(dict(full_path=path, media_type=media_type))

    def close(self):

        manifest = BytesIO()
        manifest_xml(manifest, self.files)
        manifest.seek(0)
        self.zf.writestr('META-INF/manifest.xml', manifest.getvalue())
        self.zf.writestr('mimetype', 'application/vnd.oasis.opendocument.text')

        self.zf.close()


def manifest_xml(f, files):
    from xml.sax.saxutils import XMLGenerator
    xml = XMLGenerator(f, 'utf-8')
    xml.startDocument()

    uri = 'urn:oasis:names:tc:opendocument:xmlns:manifest:1.0'
    prefix = 'manifest'
    xml.startPrefixMapping(prefix, uri)

    def startElement(name, attrs):
        attrs = dict(((uri, n), v) for n, v in attrs.items())
        xml.startElementNS((uri, name), prefix + ':' + name, attrs)

    def endElement(name):
        xml.endElementNS((uri, name), prefix + ':' + name)

    def file_entry(full_path, media_type, **kwargs):
        attrs = {'media-type': media_type, 'full-path': full_path}
        attrs.update(dict((n.replace('_', '-'), v)
                          for n, v in kwargs.items()))
        startElement('file-entry', attrs)
        endElement('file-entry')

    startElement('manifest', dict(version='1.2'))
    file_entry('/', 'application/vnd.oasis.opendocument.text', version='1.2')
    for e in files:
        e = dict(e)
        full_path = e.pop('full_path')
        media_type = e.pop('media_type', 'application/octet-stream')
        file_entry(full_path, media_type)
    endElement('manifest')

    xml.endPrefixMapping(prefix)
    xml.endDocument()


def manifest_rdf(f):
    f.write(b'''<?xml version="1.0" encoding="utf-8"?>
<rdf:RDF
    xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
    xmlns:pkg="http://docs.oasis-open.org/ns/office/1.2/meta/pkg#"
    xmlns:odf="http://docs.oasis-open.org/ns/office/1.2/meta/odf#">
    <pkg:Document rdf:about="">
        <pkg:hasPart rdf:resource="content.xml"/>
        <pkg:hasPart rdf:resource="styles.xml"/>
    </pkg:Document>
    <odf:ContentFile rdf:about="content.xml"/>
    <odf:StylesFile rdf:about="styles.xml"/>
</rdf:RDF>''')


def main():
    from docopt import docopt

    from . import __version__ as version
    from .proc import rest_to_docopt
    from .proc import init_logger
    from .dataio import ParseError
    from .errors import InvalidHwp5FileError
    from .proc import init_with_environ
    from .utils import make_open_dest_file
    from .xmlmodel import Hwp5File

    doc = rest_to_docopt(__doc__)
    args = docopt(doc, version=version)
    init_logger(args)
    init_with_environ()

    hwp5path = args['<hwp5file>']

    odt_transform = ODTTransform()

    open_dest = make_open_dest_file(args['--output'])
    if args['--document']:
        odt_transform.embedbin = not args['--no-embed-image']
        transform = odt_transform.transform_hwp5_to_single_document
        open_dest = wrap_for_xml(open_dest)
    elif args['--styles']:
        odt_transform.embedbin = args['--embed-image']
        transform = odt_transform.transform_hwp5_to_styles
        open_dest = wrap_for_xml(open_dest)
    elif args['--content']:
        odt_transform.embedbin = args['--embed-image']
        transform = odt_transform.transform_hwp5_to_content
        open_dest = wrap_for_xml(open_dest)
    else:
        odt_transform.embedbin = args['--embed-image']
        transform = odt_transform.transform_hwp5_to_package
        dest_path = args['--output']
        dest_path = dest_path or replace_ext(hwp5path, '.odt')
        open_dest = partial(open_odtpkg, dest_path)

    try:
        with closing(Hwp5File(hwp5path)) as hwp5file:
            with open_dest() as dest:
                transform(hwp5file, dest)
    except ParseError as e:
        e.print_to_logger(logger)
    except InvalidHwp5FileError as e:
        logger.error('%s', e)
        sys.exit(1)


def replace_ext(path, ext):
    name = os.path.basename(path)
    root = os.path.splitext(name)[0]
    return root + ext


@contextmanager
def open_odtpkg(path):
    odtpkg = ODTPackage(path)
    with closing(odtpkg):
        yield odtpkg


def wrap_for_xml(open_dest):
    from .utils import wrap_open_dest_for_tty
    from .utils import pager
    from .utils import syntaxhighlight
    from .utils import xmllint
    return wrap_open_dest_for_tty(open_dest, [
        pager(),
        syntaxhighlight('application/xml'),
        xmllint(format=True),
    ])



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel.controls.common_controls import CommonControl
from hwp5.binmodel.controlchar import CHID
from hwp5.binmodel.tagid61_table import TableBody


class TableControl(CommonControl):
    chid = CHID.TBL

    def on_child(cls, attributes, context, child):
        child_context, child_model = child
        if child_model['type'] is TableBody:
            # referenced in ListHeader parsing
            context['seen_table_body'] = True
    on_child = classmethod(on_child)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel.controlchar import CHID
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import UINT32
from hwp5.dataio import UINT16
from hwp5.dataio import WCHAR
from hwp5.binmodel.controls._shared import Control


class NumberingControl(Control):
    Kind = Enum(PAGE=0, FOOTNOTE=1, ENDNOTE=2, PICTURE=3, TABLE=4, EQUATION=5)
    Flags = Flags(UINT32,
                  0, 3, Kind, 'kind',
                  4, 11, 'footnoteshape',
                  12, 'superscript')

    def attributes(cls):
        yield cls.Flags, 'flags',
        yield UINT16, 'number',
    attributes = classmethod(attributes)


class AutoNumbering(NumberingControl):
    ''' 4.2.10.5. 자동 번호 '''
    chid = CHID.ATNO

    def attributes(cls):
        yield WCHAR, 'usersymbol',
        yield WCHAR, 'prefix',
        yield WCHAR, 'suffix',
    attributes = classmethod(attributes)

    def __unicode__(self):
        prefix = u''
        suffix = u''
        if self.flags.kind == self.Kind.FOOTNOTE:
            if self.suffix != u'\x00':
                suffix = self.suffix
        return prefix + unicode(self.number) + suffix


class NewNumbering(NumberingControl):
    ''' 4.2.10.6. 새 번호 지정 '''
    chid = CHID.NWNO



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import os.path
import sys

from ..errors import InvalidOleStorageError


PY3 = sys.version_info.major == 3
if PY3:
    basestring = str


def is_enabled():
    try:
        from org.apache.poi.poifs.filesystem import POIFSFileSystem
        POIFSFileSystem  # silencing
        return True
    except ImportError:
        return False


class OleStorage(object):
    ''' Create an OleStorage instance.

    :param olefile: an OLE2 Compound Binary File.
    :raises: `InvalidOleStorageError` when `olefile` is not valid OLE2 format.
    '''

    def __init__(self, olefile):
        from java.io import FileInputStream
        from java.io import IOException
        from org.apache.poi.poifs.filesystem import POIFSFileSystem
        from org.apache.poi.poifs.filesystem import DirectoryEntry

        if isinstance(olefile, basestring):
            path = os.path.abspath(olefile)
            fis = FileInputStream(path)
            try:
                fs = POIFSFileSystem(fis)
            except IOException as e:
                raise InvalidOleStorageError(e.getMessage())
            entry = fs.getRoot()
        elif isinstance(olefile, DirectoryEntry):
            entry = olefile
        else:
            raise ValueError('invalid olefile')

        self.entry = entry

    def __iter__(self):
        return (entry.getName() for entry in self.entry.getEntries())

    def __getitem__(self, name):
        from java.io import FileNotFoundException
        try:
            entry = self.entry.getEntry(name)
        except FileNotFoundException:
            raise KeyError('%s not found' % name)

        if entry.directoryEntry:
            return OleStorage(entry)
        elif entry.documentEntry:
            return OleStream(entry)
        else:
            raise KeyError('%s is invalid' % name)

    def close(self):
        return


class OleStream(object):

    def __init__(self, entry):
        self.entry = entry

    def open(self):
        from org.apache.poi.poifs.filesystem import DocumentInputStream
        dis = DocumentInputStream(self.entry)
        return FileFromDocumentInputStream(dis)


class FileFromDocumentInputStream(object):

    def __init__(self, dis):
        self.dis = dis
        self.size = dis.available()
        dis.mark(0)

    def read(self, size=None):
        import jarray
        dis = self.dis
        available = dis.available()
        if size is None:
            size = available
        elif size > available:
            size = available
        bytes = jarray.zeros(size, 'b')
        n_read = dis.read(bytes)
        data = bytes.tostring()
        if n_read < size:
            return data[:n_read]
        return data

    def seek(self, offset, whence=0):
        dis = self.dis
        if whence == 0:
            dis.reset()
            dis.skip(offset)
        elif whence == 1:
            dis.skip(offset)
        elif whence == 2:
            dis.reset()
            dis.skip(self.size - offset)
        else:
            raise ValueError('invalid whence: %s', whence)

    def tell(self):
        return self.size - self.dis.available()

    def close(self):
        return self.dis.close()



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_BIN_DATA
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import UINT16
from hwp5.dataio import Struct
from hwp5.dataio import BSTR
from hwp5.dataio import SelectiveType
from hwp5.dataio import ref_member_flag
from hwp5.binmodel._shared import BinStorageId


class BinDataLink(Struct):
    def attributes():
        yield BSTR, 'abspath'
        yield BSTR, 'relpath'
    attributes = staticmethod(attributes)


class BinDataEmbedding(Struct):
    def attributes():
        yield BinStorageId, 'storage_id'
        yield BSTR, 'ext'
    attributes = staticmethod(attributes)


class BinDataStorage(Struct):
    def attributes():
        yield BinStorageId, 'storage_id'
    attributes = staticmethod(attributes)


class BinData(RecordModel):
    ''' 4.1.3. 바이너리 데이터 '''

    tagid = HWPTAG_BIN_DATA

    # 표 13 바이너리 데이터 속성
    StorageType = Enum(LINK=0, EMBEDDING=1, STORAGE=2)
    CompressionType = Enum(STORAGE_DEFAULT=0, YES=1, NO=2)
    AccessState = Enum(NEVER=0, OK=1, FAILED=2, FAILED_IGNORED=3)
    Flags = Flags(UINT16,
                  0, 3, StorageType, 'storage',
                  4, 5, CompressionType, 'compression',
                  16, 17, AccessState, 'access')

    def attributes(cls):
        ''' 표 12 바이너리 데이터 '''
        yield cls.Flags, 'flags'
        yield (SelectiveType(ref_member_flag('flags', 'storage'),
                             {cls.StorageType.LINK: BinDataLink,
                              cls.StorageType.EMBEDDING: BinDataEmbedding,
                              cls.StorageType.STORAGE: BinDataStorage}),
               'bindata')
    attributes = classmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import codecs
import zlib  # this codec needs the optional zlib module !

_wbits = -15


def zlib_raw_encode(input, errors='strict'):
    assert errors == 'strict'
    output = zlib.compress(input)[2:-4]
    return (output, len(input))


def zlib_raw_decode(input, errors='strict'):
    assert errors == 'strict'
    output = zlib.decompress(input, _wbits)
    return (output, len(input))


class Codec(codecs.Codec):

    def encode(self, input, errors='strict'):
        return zlib_raw_encode(input, errors)

    def decode(self, input, errors='strict'):
        return zlib_raw_decode(input, errors)


class IncrementalEncoder(codecs.IncrementalEncoder):
    def __init__(self, errors='strict'):
        assert errors == 'strict'
        self.errors = errors
        self.compressobj = zlib.compressobj()
        self.initial = True

    def encode(self, input, final=False):
        c = self.compressobj.compress(input)
        if self.initial:
            c = c[2:]
            self.initial = False
        if final:
            c += self.compressobj.flush()[:-4]
        return c

    def reset(self):
        self.compressobj = zlib.compressobj()


class IncrementalDecoder(codecs.IncrementalDecoder):
    def __init__(self, errors='strict'):
        assert errors == 'strict'
        self.errors = errors
        self.decompressobj = zlib.decompressobj(_wbits)

    def decode(self, input, final=False):
        if final:
            if len(input) > 0:
                d = self.decompressobj.decompress(input)
            else:
                d = b''
            return d + self.decompressobj.flush()
        else:
            return self.decompressobj.decompress(input)

    def reset(self):
        self.decompressobj = zlib.decompressobj(_wbits)


class StreamWriter(object):
    def __init__(self, stream, errors='strict'):
        assert errors == 'strict'
        self.stream = stream
        self.encoder = IncrementalEncoder(errors)

    def write(self, data):
        raise NotImplementedError


class StreamReader(object):
    def __init__(self, stream, errors='strict'):
        assert errors == 'strict'
        self.stream = stream
        self.decoder = IncrementalDecoder(errors)
        self.buffer = b''
        self.offset = 0

    def read(self, size=-1):
        if size < 0:
            c = self.stream.read()
            d = self.buffer + self.decoder.decode(c, True)
            self.buffer = b''
            self.offset += len(d)
            return d

        final = False
        while True:
            if size <= len(self.buffer):
                d = self.buffer[:size]
                self.buffer = self.buffer[size:]
                self.offset += size
                return d

            if final:
                d = self.buffer
                self.buffer = b''
                self.offset += len(d)
                return d

            c = self.stream.read(8196)
            final = len(c) < 8196 or len(c)
            self.buffer += self.decoder.decode(c, final)

    def tell(self):
        return self.offset


_codecinfo = codecs.CodecInfo(
    name='zlib_raw',
    encode=zlib_raw_encode,
    decode=zlib_raw_decode,
    incrementalencoder=IncrementalEncoder,
    incrementaldecoder=IncrementalDecoder,
    streamreader=StreamReader,
    streamwriter=StreamWriter,
)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_DOC_DATA


class DocData(RecordModel):
    ''' 4.1.12. 문서 임의의 데이터 '''
    tagid = HWPTAG_DOC_DATA

    # TODO



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from itertools import islice
import json
import struct

from . import dataio
from . import filestructure
from .dataio import dumpbytes
from .dataio import Eof
from .dataio import UINT32
from .tagids import HWPTAG_BEGIN
from .tagids import tagnames
from .utils import JsonObjects


def tagname(tagid):
    return tagnames.get(tagid, 'HWPTAG%d' % (tagid - HWPTAG_BEGIN))


def Record(tagid, level, payload, size=None, seqno=None):
    if size is None:
        size = len(payload)
    d = dict(tagid=tagid, tagname=tagname(tagid), level=level,
             size=size, payload=payload)
    if seqno is not None:
        d['seqno'] = seqno
    return d


def decode_record_header(f):
    try:
        # TagID, Level, Size
        rechdr = UINT32.read(f)
        tagid = rechdr & 0x3ff
        level = (rechdr >> 10) & 0x3ff
        size = (rechdr >> 20) & 0xfff
        if size == 0xfff:
            size = UINT32.read(f)
        return (tagid, level, size)
    except Eof:
        return None


def encode_record_header(rec):
    size = len(rec['payload'])
    level = rec['level']
    tagid = rec['tagid']
    if size < 0xfff:
        hdr = (size << 20) | (level << 10) | tagid
        return struct.pack('<I', hdr)
    else:
        hdr = (0xfff << 20) | (level << 10) | tagid
        return struct.pack('<II', hdr, size)


def read_record(f, seqno):
    header = decode_record_header(f)
    if header is None:
        return
    tagid, level, size = header
    payload = dataio.readn(f, size)
    return Record(tagid, level, payload, size, seqno)


def dump_record(f, record):
    hdr = encode_record_header(record)
    f.write(hdr)
    f.write(record['payload'])


def read_records(f):
    seqno = 0
    while True:
        record = read_record(f, seqno)
        if record:
            yield record
        else:
            return
        seqno += 1


def link_records(records):
    prev = None
    for rec in records:
        if prev is not None:
            if rec['level'] == prev['level']:
                rec['sister'] = prev
                rec['parent'] = prev.get('parent')
            elif rec['level'] == prev['level'] + 1:
                rec['parent'] = prev
        yield rec
        prev = rec


def record_to_json(record, *args, **kwargs):
    ''' convert a record to json '''
    record['payload'] = list(dumpbytes(record['payload']))
    return json.dumps(record, *args, **kwargs)


def nth(iterable, n, default=None):
    try:
        return next(islice(iterable, n, None))
    except StopIteration:
        return default


def group_records_by_toplevel(records, group_as_list=True):
    ''' group records by top-level trees and return iterable of the groups
    '''
    context = dict()

    try:
        context['top'] = next(records)
    except StopIteration:
        return

    def records_in_a_tree():
        yield context.pop('top')

        for record in records:
            if record['level'] == 0:
                context['top'] = record
                return
            yield record

    while 'top' in context:
        group = records_in_a_tree()
        if group_as_list:
            group = list(group)
        yield group


class RecordStream(filestructure.VersionSensitiveItem):

    def records(self, **kwargs):
        records = read_records(self.open())
        if 'range' in kwargs:
            range = kwargs['range']
            records = islice(records, range[0], range[1])
        elif 'treegroup' in kwargs:
            groups = group_records_by_toplevel(records, group_as_list=True)
            records = nth(groups, kwargs['treegroup'])
        return records

    def record(self, idx):
        ''' get the record at `idx' '''
        return nth(self.records(), idx)

    def records_json(self, **kwargs):
        records = self.records(**kwargs)
        return JsonObjects(records, record_to_json)

    def records_treegrouped(self, group_as_list=True):
        ''' group records by top-level trees and return iterable of the groups
        '''
        records = self.records()
        return group_records_by_toplevel(records, group_as_list)

    def records_treegroup(self, n):
        ''' returns list of records in `n'th top-level tree '''
        groups = self.records_treegrouped()
        return nth(groups, n)

    def other_formats(self):
        return {'.records': self.records_json().open}


class Sections(filestructure.Sections):

    section_class = RecordStream


class Hwp5File(filestructure.Hwp5File):
    ''' Hwp5File for 'rec' layer
    '''

    docinfo_class = RecordStream
    bodytext_class = Sections



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from io import BytesIO
import codecs
import zlib

from .utils import GeneratorReader


class ZLibIncrementalDecoder(codecs.IncrementalDecoder):
    def __init__(self, errors='strict', wbits=15):
        assert errors == 'strict'
        self.errors = errors
        self.wbits = wbits
        self.reset()

    def decode(self, input, final=False):
        c = self.decompressobj.decompress(input)
        if final:
            c += self.decompressobj.flush()
        return c

    def reset(self):
        self.decompressobj = zlib.decompressobj(self.wbits)


def decompress_gen(source, bufsize=4096):
    dec = ZLibIncrementalDecoder(wbits=-15)
    exausted = False
    while not exausted:
        input = source.read(bufsize)
        if len(input) < bufsize:
            exausted = True
        yield dec.decode(input, exausted)


def decompress_experimental(source, bufsize=4096):
    ''' decompress inputstream

        stream: a file-like readable
        returns a file-like readable
    '''
    return GeneratorReader(decompress_gen(source, bufsize))


def decompress(stream):
    ''' decompress inputstream

        stream: a file-like readable
        returns a file-like readable
    '''
    return BytesIO(zlib.decompress(stream.read(), -15))  # without gzip header



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_PARA_RANGE_TAG
from hwp5.dataio import UINT32
from hwp5.dataio import Flags
from hwp5.dataio import Struct
from hwp5.dataio import X_ARRAY
from hwp5.binmodel._shared import ref_parent_member


class RangeTag(Struct):
    ''' 표 58 문단의 영역 태그 '''

    Tag = Flags(UINT32,
                0, 23, 'data',
                24, 31, 'kind')

    @classmethod
    def attributes(cls):
        yield UINT32, 'start'
        yield UINT32, 'end'
        yield cls.Tag, 'tag'


class ParaRangeTag(RecordModel):
    ''' 4.2.5. 문단의 영역 태그 '''

    tagid = HWPTAG_PARA_RANGE_TAG

    @staticmethod
    def attributes():
        yield dict(name='range_tags',
                   type=X_ARRAY(RangeTag, ref_parent_member('rangetags')))



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_SHAPE_COMPONENT_OLE
from hwp5.dataio import Flags
from hwp5.dataio import UINT32
from hwp5.dataio import INT32
from hwp5.binmodel._shared import BinStorageId
from hwp5.binmodel._shared import BorderLine


class ShapeOLE(RecordModel):
    ''' 4.2.9.5 OLE 개체 '''

    tagid = HWPTAG_SHAPE_COMPONENT_OLE

    Flags = Flags(UINT32,
                  0, 7, 'dvaspect',
                  8, 'moniker',
                  # baseline:
                  #  0 means defaut (85%)
                  #  1 means 0%
                  #  101 means 100%
                  9, 15, 'baseline')

    @classmethod
    def attributes(cls):
        yield cls.Flags, 'flags'
        yield INT32, 'extent_x'
        yield INT32, 'extent_y'
        yield BinStorageId, 'storage_id'
        yield BorderLine, 'border'



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''Generate HWPv5 Binary Spec Document

Usage::

    hwp5spec xml [--loglevel=<loglevel>]
    hwp5spec -h | --help
    hwp5spec --version

Options::

    -h --help       Show this screen
    --version       Show version
    --loglevel=<loglevel>   Set log level [default: warning]
'''

import logging
import xml.etree.ElementTree as ET


logger = logging.getLogger(__name__)


def define_enum_type(enum_type):
    attrs = dict(name=enum_type.__name__)
    if enum_type.scoping_struct:
        attrs['scope'] = enum_type.scoping_struct.__name__
    elem = ET.Element('EnumType', attrs)
    value_names = list((e, e.name) for e in enum_type.instances)
    value_names.sort()
    for value, name in value_names:
        item = ET.Element('item', dict(name=name, value=str(value)))
        elem.append(item)
    return elem


def define_bitfield(bitgroup_name, bitgroup_desc):
    attrs = dict(name=bitgroup_name,
                 lsb=str(bitgroup_desc.lsb),
                 msb=str(bitgroup_desc.msb))
    elem = ET.Element('BitField', attrs)
    elem.append(reference_type(bitgroup_desc.valuetype))
    return elem


def define_flags_type(flags_type):
    elem = ET.Element('FlagsType')
    from hwp5.dataio import BitGroupDescriptor
    base = ET.SubElement(elem, 'base')
    base.append(reference_type(flags_type.basetype))
    bitgroups = flags_type.__dict__.items()
    bitgroups = ((v.lsb, (k, v)) for k, v in bitgroups
                 if isinstance(v, BitGroupDescriptor))
    bitgroups = list(bitgroups)
    bitgroups.sort()
    bitgroups = reversed(bitgroups)
    bitgroups = ((k, v) for lsb, (k, v) in bitgroups)
    bitgroups = (define_bitfield(k, v) for k, v in bitgroups)
    for bitgroup in bitgroups:
        elem.append(bitgroup)
    return elem


def define_fixed_array_type(array_type):
    attrs = dict()
    attrs['size'] = str(array_type.size)
    elem = ET.Element('FixedArrayType', attrs)
    item_type_elem = ET.SubElement(elem, 'item-type')
    item_type_elem.append(reference_type(array_type.itemtype))
    return elem


def define_variable_length_array_type(array_type):
    elem = ET.Element('VariableLengthArrayType')
    count_type_elem = ET.SubElement(elem, 'count-type')
    count_type_elem.append(reference_type(array_type.counttype))
    item_type_elem = ET.SubElement(elem, 'item-type')
    item_type_elem.append(reference_type(array_type.itemtype))
    return elem


def define_x_array_type(t):
    elem = ET.Element('XArrayType', dict(size=t.count_reference.__doc__))
    item_type_elem = ET.SubElement(elem, 'item-type')
    item_type_elem.append(reference_type(t.itemtype))
    return elem


def define_selective_type(t):
    elem = ET.Element('SelectiveType',
                      dict(selector=t.selector_reference.__doc__))
    for k, v in t.selections.items():
        sel = ET.SubElement(elem, 'selection',
                            dict(when=make_condition_value(k)))
        sel.append(reference_type(v))
    return elem


def reference_type(t):
    attrs = dict()
    attrs['name'] = t.__name__
    attrs['meta'] = type(t).__name__
    elem = ET.Element('type-ref', attrs)

    from hwp5.dataio import EnumType
    from hwp5.dataio import FlagsType
    from hwp5.dataio import FixedArrayType
    from hwp5.dataio import X_ARRAY
    from hwp5.dataio import VariableLengthArrayType
    from hwp5.dataio import SelectiveType
    if isinstance(t, EnumType):
        if t.scoping_struct:
            elem.attrib['scope'] = t.scoping_struct.__name__
    elif isinstance(t, FlagsType):
        elem.append(define_flags_type(t))
    elif isinstance(t, FixedArrayType):
        elem.append(define_fixed_array_type(t))
    elif isinstance(t, X_ARRAY):
        elem.append(define_x_array_type(t))
    elif isinstance(t, VariableLengthArrayType):
        elem.append(define_variable_length_array_type(t))
    elif isinstance(t, SelectiveType):
        elem.append(define_selective_type(t))
    return elem


def referenced_types_by_member(member):
    t = member.get('type')
    if t:
        yield t
        for x in direct_referenced_types(t):
            yield x


def define_member(struct_type, member):
    attrs = dict(name=member['name'])

    version = member.get('version')
    if version:
        version = '.'.join(str(x) for x in version)
        attrs['version'] = version

    elem = ET.Element('member', attrs)

    t = member.get('type')
    if t:
        elem.append(reference_type(t))

    condition = member.get('condition')
    if condition:
        condition = condition.__doc__ or condition.__name__ or ''
        condition = condition.strip()
        condition_elem = ET.Element('condition')
        condition_elem.text = condition
        elem.append(condition_elem)

    return elem


def direct_referenced_types(t):
    from hwp5.dataio import FlagsType
    from hwp5.dataio import FixedArrayType
    from hwp5.dataio import X_ARRAY
    from hwp5.dataio import VariableLengthArrayType
    from hwp5.dataio import StructType
    from hwp5.dataio import SelectiveType
    if isinstance(t, FlagsType):
        for k, desc in t.bitfields.items():
            yield desc.valuetype
    elif isinstance(t, FixedArrayType):
        yield t.itemtype
    elif isinstance(t, X_ARRAY):
        yield t.itemtype
    elif isinstance(t, VariableLengthArrayType):
        yield t.counttype
        yield t.itemtype
    elif isinstance(t, StructType):
        if 'members' in t.__dict__:
            for member in t.members:
                for x in referenced_types_by_member(member):
                    yield x
    elif isinstance(t, SelectiveType):
        for selection in t.selections.values():
            yield selection


def referenced_types_by_struct_type(t):
    if 'members' in t.__dict__:
        for member in t.members:
            for x in referenced_types_by_member(member):
                yield x


def extension_sort_key(cls):
    import inspect
    key = inspect.getmro(cls)
    key = list(x.__name__ for x in key)
    key = tuple(reversed(key))
    return key


def sort_extensions(extension_types):
    extension_types = extension_types.items()
    extension_types = list((extension_sort_key(cls), (k, cls))
                           for k, cls in extension_types)
    extension_types.sort()
    extension_types = ((k, cls) for sort_key, (k, cls) in extension_types)
    return extension_types


def extensions_of_tag_model(tag_model):
    extension_types = getattr(tag_model, 'extension_types', None)
    if extension_types:
        extension_types = sort_extensions(extension_types)
        key_condition = getattr(tag_model, 'get_extension_key', None)
        key_condition = key_condition.__doc__.strip()
        for key, extension_type in extension_types:
            yield (key_condition, key), extension_type


def define_struct_type(t):
    elem = ET.Element('StructType',
                      dict(name=t.__name__))
    for extend in get_extends(t):
        elem.append(define_extends(extend))

    if 'members' in t.__dict__:
        for member in t.members:
            elem.append(define_member(t, member))
    return elem


def define_tag_model(tag_id):
    from hwp5.tagids import tagnames
    from hwp5.binmodel import tag_models
    tag_name = tagnames[tag_id]
    tag_model = tag_models[tag_id]
    elem = ET.Element('TagModel',
                      dict(tag_id=str(tag_id),
                           name=tag_name))
    elem.append(define_base_type(tag_model))
    for (name, value), extension_type in extensions_of_tag_model(tag_model):
        elem.append(define_extension(extension_type,
                                     tag_model,
                                     name,
                                     value))
    return elem


def define_base_type(t):
    elem = ET.Element('base', dict(name=t.__name__))
    return elem


def make_condition_value(value):
    from hwp5.dataio import EnumType
    if isinstance(value, tuple):
        value = tuple(make_condition_value(v) for v in value)
        return '('+', '.join(value)+')'
    elif isinstance(type(value), EnumType):
        return repr(value)
    elif isinstance(value, type):
        return value.__name__
    else:
        return str(value)


def define_extension(t, up_to_type, name, value):
    attrs = dict(name=t.__name__)
    elem = ET.Element('extension', attrs)
    condition = ET.Element('condition')
    condition.text = name + ' == ' + make_condition_value(value)
    elem.append(condition)

    for extend in get_extends(t, up_to_type):
        elem.append(define_extends(extend))

    if 'members' in t.__dict__:
        for member in t.members:
            elem.append(define_member(t, member))
    return elem


def get_extends(t, up_to_type=None):
    def take_up_to(up_to_type, mro):
        for t in mro:
            yield t
            if t is up_to_type:
                return
    from itertools import takewhile

    import inspect
    mro = inspect.getmro(t)
    mro = mro[1:]  # exclude self
    # mro = take_up_to(up_to_type, mro)
    mro = takewhile(lambda cls: cls is not up_to_type, mro)
    mro = (t for t in mro if 'members' in t.__dict__)
    mro = list(mro)
    mro = reversed(mro)
    return mro


def define_extends(t):
    attrs = dict(name=t.__name__)
    elem = ET.Element('extends', attrs)
    return elem


def define_primitive_type(t):
    attrs = dict(name=t.__name__)
    fixed_size = getattr(t, 'fixed_size', None)
    if fixed_size:
        attrs['size'] = str(fixed_size)

    elem = ET.Element('PrimitiveType', attrs)

    binfmt = getattr(t, 'binfmt', None)
    if binfmt:
        binfmt_elem = ET.Element('binfmt')
        binfmt_elem.text = binfmt
        elem.append(binfmt_elem)
    return elem


def main():
    from docopt import docopt
    from hwp5 import __version__
    from hwp5.proc import rest_to_docopt

    doc = rest_to_docopt(__doc__)
    args = docopt(doc, version=__version__)

    if '--loglevel' in args:
        loglevel = args['--loglevel'].lower()
        loglevel = dict(error=logging.ERROR,
                        warning=logging.WARNING,
                        info=logging.INFO,
                        debug=logging.DEBUG).get(loglevel, logging.WARNING)
        logger.setLevel(loglevel)
        logger.addHandler(logging.StreamHandler())

    from hwp5 import binmodel
    import sys

    enum_types = set()
    extensions = set()
    struct_types = set()
    primitive_types = set()

    root = ET.Element('binspec', dict(version=__version__))
    for tag_id, tag_model in binmodel.tag_models.items():
        logger.debug('TAG_MODEL: %s', tag_model.__name__)
        root.append(define_tag_model(tag_id))
        struct_types.add(tag_model)

        from hwp5.dataio import EnumType
        from hwp5.dataio import StructType
        from hwp5.dataio import PrimitiveType
        for t in referenced_types_by_struct_type(tag_model):
            if isinstance(t, EnumType):
                enum_types.add(t)
            if isinstance(t, StructType):
                struct_types.add(t)
            if isinstance(t, PrimitiveType):
                logger.debug('- PrimitiveType: %s', t.__name__)
                primitive_types.add(t)

        for _, t in extensions_of_tag_model(tag_model):
            extensions.add(t)

    for t in extensions:
        struct_types.add(t)
        for extends in get_extends(t):
            struct_types.add(extends)

    for struct_type in struct_types:
        for t in referenced_types_by_struct_type(struct_type):
            if isinstance(t, EnumType):
                enum_types.add(t)
            if isinstance(t, PrimitiveType):
                primitive_types.add(t)

    enum_types = list((t.__name__, t) for t in enum_types)
    enum_types.sort()
    enum_types = (t for name, t in enum_types)
    for t in enum_types:
        root.append(define_enum_type(t))

    struct_types = list((t.__name__, t) for t in struct_types)
    struct_types.sort()
    struct_types = (t for name, t in struct_types)
    for t in struct_types:
        root.append(define_struct_type(t))

    primitive_types = list((t.__name__, t) for t in primitive_types)
    primitive_types.sort()
    primitive_types = (t for name, t in primitive_types)
    for t in primitive_types:
        root.append(define_primitive_type(t))

    doc = ET.ElementTree(root)
    doc.write(sys.stdout, 'utf-8')



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from subprocess import CalledProcessError
from subprocess import Popen
import logging
import subprocess


logger = logging.getLogger(__name__)

executable = 'xsltproc'
enabled = None


def xslt_reachable():
    args = [executable, '--version']
    try:
        subprocess.check_output(args)
    except OSError:
        return False
    except CalledProcessError:
        return False
    except Exception as e:
        logger.exception(e)
        return False
    else:
        return True


def is_enabled():
    global enabled
    if enabled is None:
        enabled = xslt_reachable()
    return enabled


def enable():
    global enabled
    enabled = True


def disable():
    global enabled
    enabled = False


def xslt(xsl_path, inp_path, out_path):
    xslt = XSLT(xsl_path)
    return xslt.transform(inp_path, out_path)


class XSLT:

    def __init__(self, xsl_path, **params):
        self.xsl_path = xsl_path
        self.cmd = [executable]
        for name, value in params.items():
            self.cmd.extend(['--stringparam', name, value])

    def transform(self, input, output):
        '''
        >>> T.transform('input.xml', 'output.xml')
        '''
        cmd = self.cmd + ['-o', output, self.xsl_path, input]
        logger.info('%r', cmd)
        p = Popen(cmd)
        p.wait()
        if p.returncode == 0:
            return dict()
        else:
            return dict(errors=[])

    def transform_into_stream(self, input, output):
        '''
        >>> T.transform_into_stream('input.xml', sys.stdout)
        '''
        cmd = self.cmd + [self.xsl_path, input]
        logger.info('%r', cmd)
        p = Popen(cmd, stdout=output)
        p.wait()
        if p.returncode == 0:
            return dict()
        else:
            return dict(errors=[])


def xslt_compile(xsl_path, **params):
    xslt = XSLT(xsl_path, **params)
    return xslt.transform_into_stream



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
''' Print the record structure.

Usage::

    hwp5proc records [--simple | --json | --raw | --raw-header | --raw-payload]
                     [--treegroup=<treegroup> | --range=<range>]
                     [--loglevel=<loglevel>] [--logfile=<logfile>]
                     <hwp5file> <record-stream>
    hwp5proc records [--simple | --json | --raw | --raw-header | --raw-payload]
                     [--treegroup=<treegroup> | --range=<range>]
                     [--loglevel=<loglevel>] [--logfile=<logfile>]
    hwp5proc records --help

Options::

    -h --help               Show this screen
       --loglevel=<level>   Set log level.
       --logfile=<file>     Set log file.

       --simple             Print records as simple tree
       --json               Print records as json
       --raw                Print records as is
       --raw-header         Print record headers as is
       --raw-payload        Print record payloads as is

       --range=<range>      Print records specified in the <range>.
       --treegroup=<treegroup>
                            Print records specified in the <treegroup>.

    <hwp5file>              HWPv5 files (*.hwp)
    <record-stream>         Record-structured internal streams.
                            (e.g. DocInfo, BodyText/*)
    <range>                 Specifies the range of the records.
                             N-M means "from the record N to M-1 (excluding M)"
                             N means just the record N
    <treegroup>             Specifies the N-th subtree of the record structure.

Example::

    $ hwp5proc records samples/sample-5017.hwp DocInfo

Example::

    $ hwp5proc records samples/sample-5017.hwp DocInfo --range=0-2

If neither <hwp5file> nor <record-stream> is specified, the record stream is
read from the standard input with an assumption that the input is in the format
version specified by -V option.

Example::

    $ hwp5proc records --raw samples/sample-5017.hwp DocInfo --range=0-2 \
> tmp.rec
    $ hwp5proc records < tmp.rec

'''
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import sys

from ..recordstream import Hwp5File
from ..recordstream import RecordStream
from ..recordstream import encode_record_header
from ..recordstream import dump_record
from ..storage import Open2Stream
from . import parse_recordstream_name


PY2 = sys.version_info.major == 2


def main(args):
    if PY2:
        stdout_text = sys.stdout
        stdout_binary = sys.stdout
    else:
        stdout_text = sys.stdout
        stdout_binary = sys.stdout.buffer

    filename = args['<hwp5file>']
    if filename:
        hwpfile = Hwp5File(filename)
        streamname = args['<record-stream>']
        stream = parse_recordstream_name(hwpfile, streamname)
    else:
        stream = RecordStream(Open2Stream(lambda: sys.stdin), None)

    opts = dict()
    rng = args['--range']
    if rng:
        rng = rng.split('-', 1)
        rng = tuple(int(x) for x in rng)
        if len(rng) == 1:
            rng = (rng[0], rng[0] + 1)
        opts['range'] = rng
    treegroup = args['--treegroup']
    if treegroup is not None:
        opts['treegroup'] = int(treegroup)

    if args['--simple']:
        for record in stream.records(**opts):
            stdout_text.write('{:04d} {} {}\n'.format(
                record['seqno'],
                '  ' * record['level'],
                record['tagname'],
            ))
    elif args['--raw']:
        for record in stream.records(**opts):
            dump_record(stdout_binary, record)
    elif args['--raw-header']:
        for record in stream.records(**opts):
            hdr = encode_record_header(record)
            stdout_binary.write(hdr)
    elif args['--raw-payload']:
        for record in stream.records(**opts):
            stdout_binary.write(record['payload'])
    else:
        stream.records_json(**opts).dump(stdout_text)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
''' List streams in the <hwp5file>.

Usage::

    hwp5proc ls [--loglevel=<loglevel>] [--logfile=<logfile>]
                [--vstreams | --ole]
                <hwp5file>
    hwp5proc ls --help

Options::

    -h --help               Show this screen
       --loglevel=<level>   Set log level.
       --logfile=<file>     Set log file.

       --vstreams           Process with virtual streams (i.e. parsed/converted
                            form of real streams)
       --ole                Treat <hwpfile> as an OLE Compound File. As a
                            result, some streams will be presented as-is. (i.e.
                            not decompressed)

Example: List without virtual streams::

    $ hwp5proc ls sample/sample-5017.hwp

    \\x05HwpSummaryInformation
    BinData/BIN0002.jpg
    BinData/BIN0002.png
    BinData/BIN0003.png
    BodyText/Section0
    DocInfo
    DocOptions/_LinkDoc
    FileHeader
    PrvImage
    PrvText
    Scripts/DefaultJScript
    Scripts/JScriptVersion

Example: List virtual streams too::

    $ hwp5proc ls --vstreams sample/sample-5017.hwp

    \\x05HwpSummaryInformation
    \\x05HwpSummaryInformation.txt
    BinData/BIN0002.jpg
    BinData/BIN0002.png
    BinData/BIN0003.png
    BodyText/Section0
    BodyText/Section0.models
    BodyText/Section0.records
    BodyText/Section0.xml
    BodyText.xml
    DocInfo
    DocInfo.models
    DocInfo.records
    DocInfo.xml
    DocOptions/_LinkDoc
    FileHeader
    FileHeader.txt
    PrvImage
    PrvText
    PrvText.utf8
    Scripts/DefaultJScript
    Scripts/JScriptVersion

'''
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from ..storage import printstorage
from . import open_hwpfile


def main(args):
    hwpfile = open_hwpfile(args)
    printstorage(hwpfile)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2018 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import
from __future__ import print_function
from distutils.command.build import build as _build
import io
import os.path
import subprocess
import sys


def setupdir(f):
    ''' Decorate f to run inside the directory where setup.py resides.
    '''
    setup_dir = os.path.dirname(os.path.abspath(__file__))

    def wrapped(*args, **kwargs):
        old_dir = os.path.abspath(os.curdir)
        try:
            os.chdir(setup_dir)
            return f(*args, **kwargs)
        finally:
            os.chdir(old_dir)

    return wrapped


@setupdir
def import_setuptools():
    try:
        import setuptools
        return setuptools
    except ImportError:
        pass

    import ez_setup
    ez_setup.use_setuptools()
    import setuptools
    return setuptools


@setupdir
def readfile(filename):
    with io.open(filename, 'r', encoding='utf-8') as f:
        return f.read()


def get_version():
    version = readfile('VERSION.txt')
    version = version.strip()
    return version


def get_long_description():
    long_description = readfile('README')
    return long_description


def get_classifiers():
    classifiers = readfile('classifiers.txt')
    classifiers = classifiers.strip()
    classifiers = classifiers.split('\n')
    classifiers = sorted(classifiers)
    return classifiers


def get_install_requires():
    requires = readfile('requirements.in')
    requires = requires.strip()
    requires = requires.split('\n')
    requires = list(requires)
    return requires


class build(_build):
    def run(self):

        #
        # compile message catalogs
        #
        domains = [
            'hwp5proc',
            'hwp5html',
            'hwp5odt',
            'hwp5txt',
            'hwp5view',
        ]
        for domain in domains:
            args = [
                sys.executable,
                __file__,
                'compile_catalog',
                '--domain={}'.format(domain)
                # ..other common options are provided by setup.cfg
            ]
            subprocess.check_call(args)

        #
        # process to normal build operations
        #
        _build.run(self)


setup_info = {

    # basic information

    'name': 'devhwp',
    'version': get_version(),
    'description': 'hwp file format parser',
    'long_description': get_long_description(),

    # authorative

    'author': 'mete0r',
    'author_email': 'mete0r@sarangbang.or.kr',
    'license': 'GNU Affero General Public License v3 or later (AGPLv3+)',
    'url': 'https://github.com/mete0r/pyhwp',

    # classifying

    'classifiers': get_classifiers(),

    'keywords': 'hwp',

    # packaging

    'setup_requires': [
        'babel',
    ],

    'packages': [
        'hwp5',
        'hwp5.binmodel',
        'hwp5.binmodel.controls',
        'hwp5.plat',
        'hwp5.plat._uno',
        'hwp5.proc',
        'hwp5.storage',
        'hwp5.transforms',
    ],
    # do not use '.'; just omit to specify setup.py directory
    'package_dir': {
        '': 'pyhwp',
    },

    'package_data': {
        'hwp5': [
            'README',
            'COPYING',
            'VERSION.txt',
            'xsl/*.xsl',
            'xsl/odt/*.xsl',
            'odf-relaxng/OpenDocument-v1.2-os-*.rng',
            'locale/*/*/*.mo',
        ],
    },

    # 'python_requires': '>=2.7, <3',

    # installation

    'zip_safe': False,

    'entry_points': {
        'console_scripts': [
            'hwp5spec=hwp5.binspec:main',
            'hwp5proc=hwp5.proc:main',
            'hwp5odt=hwp5.hwp5odt:main',
            'hwp5txt=hwp5.hwp5txt:main',
            'hwp5html=hwp5.hwp5html:main',
            'hwp5view=hwp5.hwp5view:main',
        ]
    },

    'install_requires': get_install_requires(),
}


@setupdir
def main():
    setuptools = import_setuptools()
    setup_info['cmdclass'] = {
        'build': build,
    }
    setuptools.setup(**setup_info)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel.controlchar import CHID
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import UINT32
from hwp5.dataio import BSTR
from hwp5.binmodel.controls._shared import Control


class Dutmal(Control):
    ''' 4.2.10.13. 덧말 '''
    chid = CHID.TDUT
    Position = Enum(ABOVE=0, BELOW=1, CENTER=2)
    Align = Enum(BOTH=0, LEFT=1, RIGHT=2, CENTER=3, DISTRIBUTE=4,
                 DISTRIBUTE_SPACE=5)

    def attributes(cls):
        yield BSTR, 'maintext'
        yield BSTR, 'subtext'
        yield Flags(UINT32,
                    0, 31, cls.Position, 'position'), 'position-flags'
        yield UINT32, 'fsizeratio'
        yield UINT32, 'option'
        yield UINT32, 'stylenumber'
        yield Flags(UINT32,
                    0, 31, cls.Align, 'align'), 'align-flags'
    attributes = classmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from collections import deque
from pprint import pprint
import logging
import struct
import sys

from .dataio import BSTR
from .dataio import FixedArrayType
from .dataio import FlagsType
from .dataio import ParseError
from .dataio import SelectiveType
from .dataio import StructType
from .dataio import VariableLengthArrayType
from .dataio import X_ARRAY
from .dataio import readn
from .treeop import STARTEVENT, ENDEVENT
from .treeop import iter_subevents


logger = logging.getLogger(__name__)


def bintype_map_events(bin_item):
    bin_type = bin_item['type']
    if isinstance(bin_type, StructType):
        yield STARTEVENT, bin_item
        if hasattr(bin_type, 'members'):
            for member in bin_type.members:
                for x in bintype_map_events(member):
                    yield x
        yield ENDEVENT, bin_item
    elif isinstance(bin_type, FixedArrayType):
        yield STARTEVENT, bin_item
        item = dict(type=bin_type.itemtype)
        for x in bintype_map_events(item):
            yield x
        yield ENDEVENT, bin_item
    elif isinstance(bin_type, VariableLengthArrayType):
        yield STARTEVENT, bin_item
        item = dict(type=bin_type.itemtype)
        for x in bintype_map_events(item):
            yield x
        yield ENDEVENT, bin_item
    elif isinstance(bin_type, X_ARRAY):
        yield STARTEVENT, bin_item
        item = dict(type=bin_type.itemtype)
        for x in bintype_map_events(item):
            yield x
        yield ENDEVENT, bin_item
    elif isinstance(bin_type, SelectiveType):
        yield STARTEVENT, bin_item
        for k, v in bin_type.selections.items():
            item = dict(bin_item, select_when=k, type=v)
            for x in bintype_map_events(item):
                yield x
        yield ENDEVENT, bin_item
    elif isinstance(bin_type, FlagsType):
        # TODO: this should be done in model definitions
        # bin_type: used in binary reading
        # flags_type: binary value to flags type
        bin_item['bin_type'] = bin_type.basetype
        bin_item['flags_type'] = bin_type
        yield None, bin_item
    else:
        yield None, bin_item


def filter_with_version(events, version):
    for ev, item in events:
        required_version = item.get('version')
        if required_version is not None and version < required_version:
            # just consume and skip this tree
            logger.debug('skip following: (required version: %s)',
                         required_version)
            logger.debug('  %s', (ev, item))
            if ev is STARTEVENT:
                for x in iter_subevents(events):
                    pass
            continue
        yield ev, item


def make_items_immutable(events):
    stack = []
    for ev, item in events:
        if ev is None:
            item = tuple(sorted(item.items()))
        elif ev is STARTEVENT:
            item = tuple(sorted(item.items()))
            stack.append(item)
        elif ev is ENDEVENT:
            item = stack.pop()
        yield ev, item


def compile_type_definition(bin_item):
    events = bintype_map_events(bin_item)
    events = make_items_immutable(events)
    return tuple(events)


master_typedefs = dict()


def get_compiled_typedef(type):
    if type not in master_typedefs:
        logger.info('compile typedef of %s', type)
        typedef_events = compile_type_definition(dict(type=type))
        master_typedefs[type] = typedef_events
    return master_typedefs[type]


versioned_typedefs = dict()


def get_compiled_typedef_with_version(type, version):
    if version not in versioned_typedefs:
        versioned_typedefs[version] = typedefs = dict()
    typedefs = versioned_typedefs[version]

    if type not in typedefs:
        logger.info('filter compiled typedef of %s with version %s',
                    type, version)
        typedef_events = get_compiled_typedef(type)
        events = static_to_mutable(typedef_events)
        events = filter_with_version(events, version)
        events = make_items_immutable(events)
        events = tuple(events)
        typedefs[type] = events

    return typedefs[type]


class ERROREVENT(object):
    pass


def static_to_mutable(events):
    stack = []
    for ev, item in events:
        if ev is None:
            item = dict(item)
        elif ev is STARTEVENT:
            item = dict(item)
            stack.append(item)
        elif ev is ENDEVENT:
            item = stack.pop()
        yield ev, item


def pop_subevents(events_deque):
    level = 0
    while len(events_deque) > 0:
        event, item = events_deque.popleft()
        yield event, item
        if event is STARTEVENT:
            level += 1
        elif event is ENDEVENT:
            if level > 0:
                level -= 1
            else:
                return


def resolve_typedefs(typedef_events, context):

    array_types = (X_ARRAY, VariableLengthArrayType, FixedArrayType)

    stack = []
    selective_stack = []

    events = static_to_mutable(typedef_events)
    events = deque(events)
    while len(events) > 0:
        ev, item = events.popleft()
        if isinstance(item['type'], SelectiveType):
            if ev is STARTEVENT:
                parent_struct = stack[-1]
                struct_value = parent_struct['value']
                selector_reference = item['type'].selector_reference
                select_key = selector_reference(context, struct_value)
                logger.debug('select_key: %s', select_key)
                item['select_key'] = select_key
                selective_stack.append(item)
            elif ev is ENDEVENT:
                selective_stack.pop()
            else:
                assert False
        elif 'select_when' in item:
            assert ev in (None, STARTEVENT)
            select_key = selective_stack[-1]['select_key']
            select_when = item.pop('select_when')
            if select_when != select_key:
                # just consume and skip this tree
                logger.debug('skip following: (select key %r != %r)',
                             select_key, select_when)
                logger.debug('  %s', (ev, item))
                if ev is STARTEVENT:
                    for x in pop_subevents(events):
                        logger.debug('  %s', x)
                        pass
                continue
            logger.debug('selected for: %r', select_when)
            events.appendleft((ev, item))
        elif 'condition' in item:
            assert ev in (STARTEVENT, None)
            condition = item.pop('condition')
            parent_struct = stack[-1]
            if not condition(context, parent_struct['value']):
                # just consume and skip this tree
                logger.debug('skip following: (not matched condition: %s)',
                             condition)
                logger.debug('  %s', (ev, item))
                if ev is STARTEVENT:
                    for x in pop_subevents(events):
                        logger.debug('  %s', x)
                        pass
                continue
            events.appendleft((ev, item))
        elif isinstance(item['type'], array_types) and 'count' not in item:
            assert ev is STARTEVENT

            if isinstance(item['type'], X_ARRAY):
                parent_struct = stack[-1]
                struct_value = parent_struct['value']

                count_reference = item['type'].count_reference
                count = count_reference(context, struct_value)
            elif isinstance(item['type'], VariableLengthArrayType):
                count = dict(type=item['type'].counttype, dontcollect=True)
                yield None, count
                count = count['value']
            elif isinstance(item['type'], FixedArrayType):
                count = item['type'].size
            item['count'] = count

            subevents = list(pop_subevents(events))
            endevent = subevents[-1]
            subevents = subevents[:-1]

            def clone(events):
                stack = []
                for ev, item in events:
                    if ev in (STARTEVENT, None):
                        item = dict(item)
                        if ev is STARTEVENT:
                            stack.append(item)
                    else:
                        item = stack.pop()
                    yield ev, item

            events.appendleft(endevent)
            for _ in range(0, count):
                cloned = list(clone(subevents))
                events.extendleft(reversed(cloned))
            events.appendleft((ev, item))
        else:
            if ev is STARTEVENT:
                stack.append(item)
            elif ev is ENDEVENT:
                stack.pop()
            yield ev, item


def evaluate_bin_values(events):
    for ev, item in events:
        if 'flags_type' in item:
            flags_type = item['flags_type']
            assert isinstance(flags_type, FlagsType)
            item['value'] = flags_type(item['value'])
        yield ev, item


def construct_composite_values(events):

    stack = []

    for ev, item in events:
        if ev is STARTEVENT:
            if isinstance(item['type'], StructType):
                item['value'] = dict()
            elif isinstance(item['type'], (X_ARRAY, VariableLengthArrayType,
                                           FixedArrayType)):
                item['value'] = list()
            else:
                assert False
            stack.append(item)
        elif ev in (None, ENDEVENT):
            if ev is ENDEVENT:
                item = stack.pop()
                if isinstance(item['type'], FixedArrayType):
                    item['value'] = tuple(item['value'])

            if len(stack) > 0:
                if not item.get('dontcollect', False):
                    if isinstance(stack[-1]['type'], StructType):
                        # reduce a struct member into struct value
                        stack[-1]['value'][item['name']] = item['value']
                    elif isinstance(stack[-1]['type'],
                                    (X_ARRAY,
                                     VariableLengthArrayType,
                                     FixedArrayType)):
                        stack[-1]['value'].append(item['value'])
        yield ev, item


def log_events(events, log_fn):
    for ev, item in events:
        if ev in (STARTEVENT, ENDEVENT):
            fmt = ['%s:']
            val = [ev.__name__]
        else:
            fmt = ['  %04x:']
            val = [item['bin_offset']]

        fmt.append('%s')
        val.append(item['type'].__name__)

        if 'name' in item:
            fmt.append('%r')
            val.append(str(item['name']))

        if 'value' in item and ev is None:
            fmt.append('%r')
            val.append(item['value'])

        if 'exception' in item:
            fmt.append('-- Exception: %r')
            val.append(item['exception'])

        log_fn(' '.join(fmt), *val)
        yield ev, item


def eval_typedef_events(typedef_events, context, resolve_values):
    events = static_to_mutable(typedef_events)
    events = resolve_typedefs(events, context)
    events = resolve_values(events)
    events = evaluate_bin_values(events)
    events = construct_composite_values(events)
    events = log_events(events, logger.debug)
    return events


def resolve_values_from_stream(stream):
    def resolve_values(events):
        for ev, item in events:
            if ev is None:
                item['bin_offset'] = stream.tell()
                try:
                    item['value'] = resolve_value_from_stream(item, stream)
                except Exception as e:
                    item['exception'] = e
                    ev = ERROREVENT
            yield ev, item
    return resolve_values


def resolve_value_from_stream(item, stream):
    from hwp5.binmodel import ParaTextChunks
    from hwp5.binmodel import CHID
    if 'bin_type' in item:
        item_type = item['bin_type']
    else:
        item_type = item['type']
    if hasattr(item_type, 'binfmt'):
        binfmt = item_type.binfmt
        binsize = struct.calcsize(binfmt)
        bytes = readn(stream, binsize)
        unpacked = struct.unpack(binfmt, bytes)
        return unpacked[0]
    elif item_type is CHID:
        bytes = readn(stream, 4)
        return CHID.decode(bytes)
    elif item_type is BSTR:
        return BSTR.read(stream)
    elif item_type is ParaTextChunks:
        return ParaTextChunks.read(stream)
    elif hasattr(item_type, 'fixed_size'):
        bytes = readn(stream, item_type.fixed_size)
        if hasattr(item_type, 'decode'):
            return item_type.decode(bytes)
        return bytes
    else:
        assert hasattr(item_type, 'read')
        logger.warning('%s: item type relies on its read() to resolve a value',
                       item_type.__name__)
        return item_type.read(stream)


def resolve_type_events(type, context, resolve_values):
    # get typedef events: if current version is specified in the context,
    # get version specific typedef
    if 'version' in context:
        version = context['version']
        events = get_compiled_typedef_with_version(type, version)
    else:
        events = get_compiled_typedef(type)

    # evaluate with context/stream
    return eval_typedef_events(events, context, resolve_values)


def read_type_events(type, context, stream):
    resolve_values = resolve_values_from_stream(stream)
    events = resolve_type_events(type, context, resolve_values)
    for ev, item in events:
        yield ev, item
        if ev is ERROREVENT:
            e = item['exception']
            msg = 'can\'t parse %s' % type
            pe = ParseError(msg)
            pe.cause = e
            pe.path = context.get('path')
            pe.treegroup = context.get('treegroup')
            pe.record = context.get('record')
            pe.offset = item.get('bin_offset')
            raise pe


def read_type_item(type, context, stream, binevents=None):
    if binevents is None:
        binevents = []
    try:
        binevents.extend(read_type_events(type, context, stream))
    except ParseError as e:
        e.binevents = binevents
        raise
    return binevents[-1][1]


def read_type(type, context, stream, binevents=None):
    item = read_type_item(type, context, stream, binevents)
    return item['value']


def dump_events(events):
    def prefix_level(event_prefixed_items):
        level = 0
        for ev, item in event_prefixed_items:
            if ev is STARTEVENT:
                yield level, item
                level += 1
            elif ev is ENDEVENT:
                level -= 1
            else:
                yield level, item

    def item_to_dict(events):
        for ev, item in events:
            yield ev, dict(item)

    def type_to_string(events):
        for ev, item in events:
            item['type'] = item['type'].__name__
            yield ev, item

    def condition_to_string(events):
        for ev, item in events:
            if 'condition' in item:
                item['condition'] = item['condition'].__name__
            yield ev, item

    events = item_to_dict(events)
    events = type_to_string(events)
    events = condition_to_string(events)
    for level, item in prefix_level(events):
        indents = ''
        if level > 0:
            if level > 1:
                indents = '  ' * (level - 2) + '  '
            indents += '- '
        print('{}{}'.format(indents, item))


def main():
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler())

    import hwp5.binmodel
    name = sys.argv[1]
    type = getattr(hwp5.binmodel, name)
    typedef_events = compile_type_definition(dict(type=type))
    pprint(typedef_events)

    context = {}

    def resolve_values(events):
        for ev, item in events:
            if ev is None:
                print('')
                for k, v in sorted(item.items()):
                    print('- {} : {}'.format(k, v))
                value = raw_input('>> ')
                value = eval(value)
                if isinstance(item['type'], FlagsType):
                    value = item['type'](value)
                item['value'] = value
            yield ev, item
    events = eval_typedef_events(typedef_events, context, resolve_values)
    for ev, item in events:
        print('{} {}'.format(ev, item))



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from contextlib import contextmanager
import logging

from ..errors import ImplementationNotAvailable
from ..plat import get_xslt_compile
from ..utils import hwp5_resources_path
from ..utils import mkstemp_open


logger = logging.getLogger(__name__)


class BaseTransform:

    def __init__(self, xslt_compile=None, embedbin=False):
        self.xslt_compile = xslt_compile or self.get_default_xslt_compile()
        self.embedbin = embedbin

    @classmethod
    def get_default_xslt_compile(cls):
        xslt_compile = get_xslt_compile()
        if not xslt_compile:
            raise ImplementationNotAvailable('xslt')
        return xslt_compile

    def make_transform_hwp5(self, transform_xhwp5):
        def transform_hwp5(hwp5file, output):
            with self.transformed_xhwp5_at_temp(hwp5file) as xhwp5path:
                return transform_xhwp5(xhwp5path, output)
        return transform_hwp5

    def make_xsl_transform(self, resource_path, **params):
        with hwp5_resources_path(resource_path) as xsl_path:
            return self.xslt_compile(xsl_path, **params)

    @contextmanager
    def transformed_xhwp5_at_temp(self, hwp5file):
        with mkstemp_open() as (tmp_path, f):
            hwp5file.xmlevents(embedbin=self.embedbin).dump(f)
            yield tmp_path



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_NUMBERING
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import UINT32
from hwp5.dataio import UINT16
from hwp5.dataio import INT32
from hwp5.dataio import BSTR
from hwp5.dataio import HWPUNIT16
from hwp5.dataio import Struct
from hwp5.dataio import ARRAY


class NumberingLevel(Struct):

    # 표 35 문단 머리 정보 속성
    Align = Enum(LEFT=0, CENTER=1, RIGHT=2, UNKNOWN=3)
    DistanceType = Enum(RATIO=0, VALUE=1)
    Flags = Flags(UINT32,
                  0, 1, Align, 'align',
                  2, 'auto_width',
                  3, 'auto_indent',
                  4, DistanceType, 'space_type')

    @classmethod
    def attributes(cls):
        ''' 표 34 문단 머리 정보 '''
        yield cls.Flags, 'flags'
        yield HWPUNIT16, 'width_correction'
        yield UINT16, 'space'
        yield INT32, 'charshape_id'
        yield BSTR, 'numbering_format'  # see 표 36 문단 번호 형식


class Numbering(RecordModel):
    ''' 4.1.8. 문단 번호 '''
    tagid = HWPTAG_NUMBERING

    Align = NumberingLevel.Align
    DistanceType = NumberingLevel.DistanceType
    Flags = NumberingLevel.Flags

    def attributes(cls):
        ''' 표 33 문단 번호 '''
        yield ARRAY(NumberingLevel, 7), 'levels'
        yield UINT16, 'starting_number'
        yield dict(type=ARRAY(UINT32, 7),
                   name='unknown',
                   version=(5, 0, 3, 0))
    attributes = classmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_ID_MAPPINGS
from hwp5.dataio import UINT32


class IdMappings(RecordModel):
    ''' 4.1.2. 아이디 매핑 헤더 '''

    tagid = HWPTAG_ID_MAPPINGS

    def attributes():
        ''' 표 10 아이디 매핑 헤더 '''
        yield UINT32, 'bindata',
        yield UINT32, 'ko_fonts',
        yield UINT32, 'en_fonts',
        yield UINT32, 'cn_fonts',
        yield UINT32, 'jp_fonts',
        yield UINT32, 'other_fonts',
        yield UINT32, 'symbol_fonts',
        yield UINT32, 'user_fonts',
        yield UINT32, 'borderfills',
        yield UINT32, 'charshapes',
        yield UINT32, 'tabdefs',
        yield UINT32, 'numberings',
        yield UINT32, 'bullets',
        yield UINT32, 'parashapes',
        yield UINT32, 'styles',

        # memoshapes are found from 5.0.1.7, but not in 5.0.1.6
        yield dict(type=UINT32, name='memoshapes', version=(5, 0, 1, 7))

        # TODO unknown fields:
        # followings are found from 5.0.3.2, but not in 5.0.3.1
        # but some 5.0.3.3 files do not have them:
        #   5.0.3.3/d6dfac424525298119de54410c3b22d74aa85511
        # yield dict(type=UINT32, name='unknown1', version=(5, 0, 3, 2))
        # yield dict(type=UINT32, name='unknown2', version=(5, 0, 3, 2))
    attributes = staticmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_PARA_HEADER
from hwp5.dataio import Flags
from hwp5.dataio import BYTE
from hwp5.dataio import UINT32
from hwp5.dataio import UINT16


class Paragraph(RecordModel):
    ''' 4.2.1. 문단 헤더 '''

    tagid = HWPTAG_PARA_HEADER

    # 표 54 단 나누기 종류
    SplitFlags = Flags(BYTE,
                       0, 'new_section',
                       1, 'new_columnsdef',
                       2, 'new_page',
                       3, 'new_column')
    ControlMask = Flags(UINT32,
                        2, 'unknown1',
                        11, 'control',
                        21, 'new_number')
    Flags = Flags(UINT32,
                  31, 'unknown',
                  0, 30, 'chars')

    def attributes(cls):
        ''' 표 53 문단 헤더 '''
        yield cls.Flags, 'text',
        yield cls.ControlMask, 'controlmask',
        yield UINT16, 'parashape_id',
        yield BYTE, 'style_id',
        yield cls.SplitFlags, 'split',
        yield UINT16, 'charshapes',
        yield UINT16, 'rangetags',
        yield UINT16, 'linesegs',
        yield UINT32, 'instance_id',
    attributes = classmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
''' Print parsed binary models in the specified <record-stream>.

Usage::

    hwp5proc models [--simple | --json | --format=<format> | --events]
                    [--treegroup=<treegroup> | --seqno=<seqno>]
                    [--loglevel=<loglevel>] [--logfile=<logfile>]
                    (<hwp5file> <record-stream> | -V <version>)
    hwp5proc models --help

Options::

    -h --help               Show this screen
       --loglevel=<level>   Set log level.
       --logfile=<file>     Set log file.

       --simple             Print records as simple tree
       --json               Print records as json
       --format=<format>    Print records as formatted

       --treegroup=<treegroup>
                            Print records in the <treegroup>.
                            <treegroup> specifies the N-th subtree of the
                            record structure.
       --seqno=<seqno>      Print a model of <seqno>-th record

    -V <version>, --file-format-version=<version>
                            Specifies HWPv5 file format version

    <hwp5file>              HWPv5 files (*.hwp)
    <record-stream>         Record-structured internal streams.
                            (e.g. DocInfo, BodyText/*)

Example::

    $ hwp5proc models samples/sample-5017.hwp DocInfo
    $ hwp5proc models samples/sample-5017.hwp BodyText/Section0

    $ hwp5proc models samples/sample-5017.hwp docinfo
    $ hwp5proc models samples/sample-5017.hwp bodytext/0

Example::

    $ hwp5proc models --simple samples/sample-5017.hwp bodytext/0
    $ hwp5proc models --format='%(level)s %(tagname)s\\n' \\
            samples/sample-5017.hwp bodytext/0

Example::

    $ hwp5proc models --simple --treegroup=1 samples/sample-5017.hwp bodytext/0
    $ hwp5proc models --simple --seqno=4 samples/sample-5017.hwp bodytext/0

If neither <hwp5file> nor <record-stream> is specified, the record stream is
read from the standard input with an assumption that the input is in the format
version specified by -V option.

Example::

    $ hwp5proc cat samples/sample-5017.hwp BodyText/Section0 > Section0.bin
    $ hwp5proc models -V 5.0.1.7 < Section0.bin

'''
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from itertools import islice
import sys

from ..binmodel import Hwp5File
from ..binmodel import ModelStream
from ..binmodel import RecordModel
from ..binmodel import model_to_json
from ..dataio import hexdump
from ..storage import Open2Stream
from ..treeop import ENDEVENT
from ..utils import generate_json_array
from ..utils import unicode_unescape
from . import parse_recordstream_name


PY2 = sys.version_info.major == 2


def main(args):
    stream = stream_from_args(args)
    if args['--events']:
        for event, item in stream.parse_model_events():
            type = item['type'].__name__
            if event is not None:
                if item['type'] is RecordModel:
                    record = item['record']
                    fmt = '     %s Record %s level=%s %s'
                    print(fmt % (event.__name__,
                                 record['seqno'],
                                 record['level'],
                                 record['tagname']))
                    if event is ENDEVENT:
                        leftover = item['leftover']
                        print('%04x' % leftover['offset'])
                        if len(leftover['bytes']):
                            print('')
                            print('leftover:')
                            print(hexdump(leftover['bytes']))
                        print('-' * 20)
                else:
                    print('    ', event.__name__, type, item.get('name', ''))
            else:
                offset = item['bin_offset']
                name = item.get('name', '-')
                value = item.get('value', '-')
                print('%04x' % offset, type, name, repr(value))
        return

    models_from_stream = models_from_args(args)
    models = models_from_stream(stream)

    print_models = print_models_from_args(args)
    print_models(models)


def stream_from_args(args):
    filename = args['<hwp5file>']
    if filename:
        streamname = args['<record-stream>']
        hwpfile = Hwp5File(filename)
        return parse_recordstream_name(hwpfile, streamname)
    else:
        version = args['--file-format-version'] or '5.0.0.0'
        version = version.split('.')
        version = tuple(int(x) for x in version)

        if PY2:
            stdin_binary = sys.stdin
        else:
            stdin_binary = sys.stdin.buffer

        return ModelStream(Open2Stream(lambda: stdin_binary), version)


def models_from_args(args):

    if args['--treegroup']:
        treegroup = int(args['--treegroup'])
        return lambda stream: stream.models(treegroup=treegroup)

    if args['--seqno']:
        seqno = int(args['--seqno'])
        return lambda stream: islice(stream.models(),
                                     seqno, seqno + 1)

    return lambda stream: stream.models()


def print_models_from_args(args):

    if args['--simple']:
        return print_models_with_print_model(print_model_simple)

    if args['--format']:
        fmt = args['--format']
        fmt = unicode_unescape(fmt)
        print_model = print_model_with_format(fmt)
        return print_models_with_print_model(print_model)

    return print_models_json


def print_models_json(models):
    jsonobjects = (model_to_json(model, sort_keys=True, indent=2)
                   for model in models)
    for s in generate_json_array(jsonobjects):
        sys.stdout.write(s)


def print_models_with_print_model(print_model):
    def models_printer(models):
        for model in models:
            print_model(model)
    return models_printer


def print_model_simple(model):
    sys.stdout.write('%04d ' % model['seqno'])
    sys.stdout.write(' ' * model['level'] + model['type'].__name__)
    sys.stdout.write('\n')


def print_model_with_format(fmt):
    def print_model(model):
        model = transform_model_formattable(model)
        sys.stdout.write(fmt % model)
    return print_model


def transform_model_formattable(model):
    return dict(model, type=model['type'].__name__)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from io import BytesIO
from itertools import takewhile
import json
import logging
import inspect

from .. import recordstream
from ..bintype import ERROREVENT
from ..bintype import resolve_type_events
from ..bintype import resolve_values_from_stream
from ..dataio import ParseError
from ..dataio import dumpbytes
from ..recordstream import nth
from ..tagids import tagnames
from ..treeop import STARTEVENT
from ..treeop import ENDEVENT
from ..treeop import prefix_ancestors_from_level
from ..utils import JsonObjects

from ._shared import tag_models
from ._shared import RecordModel
from ._shared import BinStorageId
from ._shared import COLORREF
from ._shared import Margin
from .controlchar import CHID
from .controlchar import ControlChar
from .tagid16_document_properties import DocumentProperties
from .tagid17_id_mappings import IdMappings
from .tagid18_bin_data import BinData
from .tagid20_border_fill import BorderFill
from .tagid19_face_name import FaceName
from .tagid21_char_shape import CharShape
from .tagid21_char_shape import LanguageStruct
from .tagid22_tab_def import TabDef
from .tagid23_numbering import Numbering
from .tagid24_bullet import Bullet
from .tagid25_para_shape import ParaShape
from .tagid26_style import Style
from .tagid27_doc_data import DocData
from .tagid28_distribute_doc_data import DistributeDocData
from .tagid30_compatible_document import CompatibleDocument
from .tagid31_layout_compatibility import LayoutCompatibility
from .tagid32_unknown import TagModel32
from .tagid50_para_header import Paragraph
from .tagid51_para_text import ParaText
from .tagid51_para_text import ParaTextChunks
from .tagid52_para_char_shape import ParaCharShape
from .tagid53_para_line_seg import ParaLineSeg
from .tagid53_para_line_seg import ParaLineSegList
from .tagid53_para_line_seg import LineSeg
from .tagid54_para_range_tag import ParaRangeTag
from .tagid55_ctrl_header import Control
from .controls.bookmark_control import BookmarkControl
from .controls.columns_def import ColumnsDef
from .controls.common_controls import CommonControl
from .controls.dutmal import Dutmal
from .controls.field import Field
from .controls.field import FieldUnknown
from .controls.field import FieldDate
from .controls.field import FieldDocDate
from .controls.field import FieldPath
from .controls.field import FieldBookmark
from .controls.field import FieldMailMerge
from .controls.field import FieldCrossRef
from .controls.field import FieldFormula
from .controls.field import FieldClickHere
from .controls.field import FieldClickHereData
from .controls.field import FieldSummary
from .controls.field import FieldUserInfo
from .controls.field import FieldHyperLink
from .controls.field import FieldMemo
from .controls.field import FieldPrivateInfoSecurity
from .controls.gshape_object_control import GShapeObjectControl
from .controls.header_footer import HeaderFooter
from .controls.header_footer import Header
from .controls.header_footer import Footer
from .controls.hidden_comment import HiddenComment
from .controls.index_marker import IndexMarker
from .controls.note import Note
from .controls.note import FootNote
from .controls.note import EndNote
from .controls.numbering import AutoNumbering
from .controls.numbering import NewNumbering
from .controls.page_hide import PageHide
from .controls.page_number_position import PageNumberPosition
from .controls.page_odd_even import PageOddEven
from .controls.section_def import SectionDef
from .controls.table_control import TableControl
from .controls.tcps_control import TCPSControl
from .tagid56_list_header import ListHeader
from .tagid56_list_header import TableCaption
from .tagid56_list_header import TableCell
from .tagid56_list_header import TextboxParagraphList
from .tagid56_list_header import HeaderParagraphList
from .tagid56_list_header import FooterParagraphList
from .tagid57_page_def import PageDef
from .tagid58_footnote_shape import FootnoteShape
from .tagid59_page_border_fill import PageBorderFill
from .tagid60_shape_component import ShapeComponent
from .tagid61_table import TableBody
from .tagid62_shape_component_line import ShapeLine
from .tagid63_shape_component_rectangle import ShapeRectangle
from .tagid64_shape_component_ellipse import ShapeEllipse
from .tagid65_shape_component_arc import ShapeArc
from .tagid66_shape_component_polygon import ShapePolygon
from .tagid67_shape_component_curve import ShapeCurve
from .tagid68_shape_component_ole import ShapeOLE
from .tagid69_shape_component_picture import ShapePicture
from .tagid70_shape_component_container import ShapeContainer
from .tagid71_ctrl_data import ControlData
from .tagid72_ctrl_eqedit import EqEdit
from .tagid74_shape_component_textart import ShapeTextArt
from .tagid75_form_object import FormObject
from .tagid76_memo_shape import MemoShape
from .tagid77_memo_list import MemoList
from .tagid78_forbidden_char import ForbiddenChar
from .tagid79_chart_data import ChartData
from .tagid99_shape_component_unknown import ShapeUnknown

# to suppress pyflake8 warning 'imported but not used'
RecordModel
BinStorageId
COLORREF
Margin
DocumentProperties
BinData
BorderFill
CharShape
LanguageStruct
TabDef
Numbering
Bullet
ParaShape
Style
DocData
DistributeDocData
CompatibleDocument
LayoutCompatibility
TagModel32
Paragraph
ParaText
ParaTextChunks
ParaCharShape
ParaLineSeg
ParaLineSegList
LineSeg
ParaRangeTag
Control
ListHeader
TableCaption
TableCell
TextboxParagraphList
PageDef
FootnoteShape
PageBorderFill
ShapeComponent
TableBody
ShapeLine
ShapeRectangle
ShapeEllipse
ShapeArc
ShapePolygon
ShapeCurve
ShapeOLE
ShapePicture
ShapeContainer
ControlData
EqEdit
ShapeTextArt
FormObject
MemoShape
MemoList
ForbiddenChar
ChartData
ShapeUnknown
CHID
ControlChar
BookmarkControl
ColumnsDef
CommonControl
Dutmal
Field
FieldUnknown
FieldDate
FieldDocDate
FieldPath
FieldBookmark
FieldMailMerge
FieldCrossRef
FieldFormula
FieldClickHere
FieldClickHereData
FieldSummary
FieldUserInfo
FieldHyperLink
FieldMemo
FieldPrivateInfoSecurity
GShapeObjectControl
HeaderFooter
Header
HeaderParagraphList
Footer
FooterParagraphList
HiddenComment
IndexMarker
Note
FootNote
EndNote
AutoNumbering
NewNumbering
PageHide
PageNumberPosition
PageOddEven
SectionDef
TableControl
TCPSControl


logger = logging.getLogger(__name__)


class UnknownTagModel(RecordModel):
    pass


class Text(object):
    pass


def _check_tag_models():
    for tagid, name in tagnames.items():
        assert tagid in tag_models, 'RecordModel for %s is missing!' % name


_check_tag_models()


def init_record_parsing_context(base, record):
    ''' Initialize a context to parse the given record

        the initializations includes followings:
        - context = dict(base)
        - context['record'] = record
        - context['stream'] = record payload stream

        :param base: the base context to be shallow-copied into the new one
        :param record: to be parsed
        :returns: new context
    '''

    return dict(base, record=record, stream=BytesIO(record['payload']))


def parse_models(context, records):
    for context, model in parse_models_intern(context, records):
        yield model


def parse_models_intern(context, records):
    context_models = ((init_record_parsing_context(context, record), record)
                      for record in records)
    context_models = parse_models_with_parent(context_models)
    for context, model in context_models:
        stream = context['stream']
        unparsed = stream.read()
        if unparsed:
            model['unparsed'] = unparsed
        yield context, model


def parse_models_with_parent(context_models):
    level_prefixed = ((model['level'], (context, model))
                      for context, model in context_models)
    root_item = (dict(), dict())
    ancestors_prefixed = prefix_ancestors_from_level(level_prefixed, root_item)
    for ancestors, (context, model) in ancestors_prefixed:
        context['parent'] = ancestors[-1]
        parse_model(context, model)
        yield context, model


def parse_model(context, model):
    ''' HWPTAG로 모델 결정 후 기본 파싱 '''

    stream = context['stream']
    context['resolve_values'] = resolve_values_from_stream(stream)
    events = resolve_model_events(context, model)
    events = raise_on_errorevent(context, events)
    model['binevents'] = list(events)

    logger.debug('model: %s', model['type'].__name__)
    logger.debug('%s', model['content'])


def raise_on_errorevent(context, events):
    binevents = list()
    for ev, item in events:
        yield ev, item
        binevents.append((ev, item))
        if ev is ERROREVENT:
            e = item['exception']
            msg = 'can\'t parse %s' % item['type']
            pe = ParseError(msg)
            pe.cause = e
            pe.path = context.get('path')
            pe.treegroup = context.get('treegroup')
            pe.record = context.get('record')
            pe.offset = item.get('bin_offset')
            pe.binevents = binevents
            raise pe


def resolve_models(context, records):
    model_contexts = (dict(context, record=record, model=dict(record))
                      for record in records)

    level_prefixed = ((context['model']['level'], context)
                      for context in model_contexts)
    root_item = {}
    ancestors_prefixed = prefix_ancestors_from_level(level_prefixed, root_item)
    for ancestors, context in ancestors_prefixed:
        parent = ancestors[-1]
        context['parent'] = parent, parent.get('model', {})

        record_frame = context['record']
        context['type'] = RecordModel
        context['name'] = record_frame['tagname']
        yield STARTEVENT, context
        for x in resolve_model_events(context, context['model']):
            yield x
        event, item = x
        context['value'] = item
        yield ENDEVENT, context


def resolve_model_events(context, model):

    resolve_values = context['resolve_values']

    model['type'] = model_type = tag_models.get(model['tagid'],
                                                UnknownTagModel)

    for ev, item in resolve_type_events(model_type, context, resolve_values):
        yield ev, item

    model['content'] = item['value']

    extension_types = getattr(model['type'], 'extension_types', None)
    if extension_types:
        key = model['type'].get_extension_key(context, model)
        extension = extension_types.get(key)
        if extension is not None:
            # 예: Control -> TableControl로 바뀌는 경우,
            # Control의 member들은 이미 읽은 상태이고
            # CommonControl, TableControl에서 각각 정의한
            # 멤버들을 읽어들여야 함
            for cls in get_extension_mro(extension, model['type']):
                extension_type_events = resolve_type_events(cls, context,
                                                            resolve_values)
                for ev, item in extension_type_events:
                    yield ev, item
                content = item['value']
                model['content'].update(content)
            model['type'] = extension

    if 'parent' in context:
        parent = context['parent']
        parent_context, parent_model = parent
        parent_type = parent_model.get('type')
        parent_content = parent_model.get('content')

        on_child = getattr(parent_type, 'on_child', None)
        if on_child:
            on_child(parent_content, parent_context, (context, model))


def get_extension_mro(cls, up_to_cls=None):
    mro = inspect.getmro(cls)
    mro = takewhile(lambda cls: cls is not up_to_cls, mro)
    mro = list(cls for cls in mro if 'attributes' in cls.__dict__)
    mro = reversed(mro)
    return mro


class ModelJsonEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, bytes):
            return obj.decode('latin1')
        return json.JSONEncoder.default(self, obj)


def model_to_json(model, *args, **kwargs):
    ''' convert a model to json '''
    kwargs['cls'] = ModelJsonEncoder
    model = dict(model)
    model['type'] = model['type'].__name__
    record = model
    record['payload'] = list(dumpbytes(record['payload']))
    if 'unparsed' in model:
        model['unparsed'] = list(dumpbytes(model['unparsed']))
    if 'binevents' in model:
        del model['binevents']
    return json.dumps(model, *args, **kwargs)


def chain_iterables(iterables):
    for iterable in iterables:
        for item in iterable:
            yield item


class ModelStream(recordstream.RecordStream):

    def models(self, **kwargs):
        # prepare binmodel parsing context
        kwargs.setdefault('version', self.version)
        try:
            kwargs.setdefault('path', self.path)
        except AttributeError:
            pass
        treegroup = kwargs.get('treegroup', None)
        if treegroup is not None:
            records = self.records_treegroup(treegroup)  # TODO: kwargs
            models = parse_models(kwargs, records)
        else:
            groups = self.models_treegrouped(**kwargs)
            models = chain_iterables(groups)
        return models

    def models_treegrouped(self, **kwargs):
        ''' iterable of iterable of the models, grouped by the top-level tree
        '''
        kwargs.setdefault('version', self.version)
        for group_idx, records in enumerate(self.records_treegrouped()):
            kwargs['treegroup'] = group_idx
            yield parse_models(kwargs, records)

    def model(self, idx):
        return nth(self.models(), idx)

    def models_json(self, **kwargs):
        models = self.models(**kwargs)
        return JsonObjects(models, model_to_json)

    def other_formats(self):
        d = super(ModelStream, self).other_formats()
        d['.models'] = self.models_json().open
        return d

    def parse_model_events(self):
        context = dict(version=self.version)

        def resolve_values_from_record(record):
            stream = BytesIO(record['payload'])
            return resolve_values_from_stream(stream)

        for group_idx, records in enumerate(self.records_treegrouped()):
            context['treegroup'] = group_idx
            for x in resolve_models(context, records):
                event, item = x
                if item['type'] is RecordModel:
                    if event is STARTEVENT:
                        record_frame = item['record']
                        stream = BytesIO(record_frame['payload'])
                        resolve_values = resolve_values_from_stream(stream)
                        item['stream'] = stream
                        item['resolve_values'] = resolve_values
                    elif event is ENDEVENT:
                        stream = item['stream']
                        item['leftover'] = {
                            'offset': stream.tell(),
                            'bytes': stream.read()
                        }
                yield x


class DocInfo(ModelStream):

    @property
    def idmappings(self):
        for model in self.models():
            if model['type'] is IdMappings:
                return model

    @property
    def facenames_by_lang(self):
        facenames = list(m for m in self.models()
                         if m['type'] is FaceName)
        languages = 'ko', 'en', 'cn', 'jp', 'other', 'symbol', 'user'
        facenames_by_lang = dict()
        offset = 0
        for lang in languages:
            n_fonts = self.idmappings['content'][lang + '_fonts']
            facenames_by_lang[lang] = facenames[offset:offset + n_fonts]
            offset += n_fonts
        return facenames_by_lang

    @property
    def charshapes(self):
        return (m for m in self.models()
                if m['type'] is CharShape)

    def get_charshape(self, charshape_id):
        return nth(self.charshapes, charshape_id)

    def charshape_lang_facename(self, charshape_id, lang):
        charshape = self.get_charshape(charshape_id)
        lang_facename_offset = charshape['content']['font_face'][lang]
        return self.facenames_by_lang[lang][lang_facename_offset]


class Sections(recordstream.Sections):

    section_class = ModelStream


class Hwp5File(recordstream.Hwp5File):

    docinfo_class = DocInfo
    bodytext_class = Sections


def create_context(file=None, **context):
    if file is not None:
        context['version'] = file.fileheader.version
    assert 'version' in context
    return context



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel.controlchar import CHID
from hwp5.dataio import UINT32
from hwp5.dataio import UINT16
from hwp5.dataio import BSTR
from hwp5.binmodel.tagid71_ctrl_data import ControlData
from hwp5.binmodel.controls._shared import Control


class BookmarkControl(Control):
    ''' 4.2.10.11. 책갈피 '''
    chid = CHID.BOKM

    def attributes():
        if False:
            yield
    attributes = staticmethod(attributes)


class BookmarkControlData(ControlData):

    parent_model_type = BookmarkControl

    def attributes():
        yield UINT32, 'unknown1'
        yield UINT32, 'unknown2'
        yield UINT16, 'unknown3'
        yield BSTR, 'name'
    attributes = staticmethod(attributes)
